# Allen Institute Software License - This software license is the 2-clause BSD
# license plus a third clause that prohibits redistribution for commercial
# purposes without further permission.
#
# Copyright 2015-2016. Allen Institute. All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice,
# this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
# this list of conditions and the following disclaimer in the documentation
# and/or other materials provided with the distribution.
#
# 3. Redistributions for commercial purposes are not permitted without the
# Allen Institute's written permission.
# For purposes of this license, commercial purposes is the incorporation of the
# Allen Institute's software into anything for which you will charge fees or
# other compensation. Contact terms@alleninstitute.org for commercial licensing
# opportunities.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
#
import simplejson as json
import logging
from ...deprecated import deprecated
from .rma_template import RmaTemplate


class GlifApi(RmaTemplate):

    _log = logging.getLogger('allensdk.api.queries.glif_api')

    NWB_FILE_TYPE = None

    rma_templates = \
        {"glif_queries": [
            {'name': 'neuronal_model_templates',
             'description': 'see name',
             'model': 'NeuronalModelTemplate',
             'num_rows': 'all',
             'count': False,
             },
            {'name': 'neuronal_models',
             'description': 'see name',
             'model': 'Specimen',
             'include': 'neuronal_models(well_known_files,neuronal_model_template,neuronal_model_runs(well_known_files))',
             'criteria':'[id$in{{ ephys_experiment_ids }}]',
             'num_rows': 'all',
             'criteria_params':['ephys_experiment_ids'],
             'count': False,
             },
            {'name': 'neuron_config',
             'description': 'see name',
             'model': 'NeuronalModel',
             'include': 'well_known_files(well_known_file_type)',
             'criteria':'[id$in{{ neuronal_model_ids }}]',
             'num_rows': 'all',
             'criteria_params':['neuronal_model_ids'],
             'count': False,
            }
        ]
        }

    def __init__(self, base_uri=None):
        super(GlifApi, self).__init__(base_uri, query_manifest=GlifApi.rma_templates)

    def get_neuronal_model_templates(self):

        return self.template_query('glif_queries',
                                   'neuronal_model_templates')

    def get_neuronal_models(self, ephys_experiment_ids=None):
        return self.template_query('glif_queries',
                                   'neuronal_models', ephys_experiment_ids=ephys_experiment_ids)

    def get_neuronal_models_by_id(self, neuronal_model_ids=None):
        return self.template_query('glif_queries',
                                   'neuron_config', neuronal_model_ids=neuronal_model_ids)

    def get_neuron_configs(self, neuronal_model_ids=None):

        data = self.template_query('glif_queries',
                                   'neuron_config', neuronal_model_ids=neuronal_model_ids)

        return_dict = {}
        for curr_config in data:
            # print curr_config
            neuron_config_url = curr_config['well_known_files'][0]['download_link']
            return_dict[curr_config['id']] = self.retrieve_parsed_json_over_http(self.api_url +
                                                                                 neuron_config_url)

        return return_dict

    @deprecated()
    def list_neuronal_models(self):
        ''' DEPRECATED Query the API for a list of all GLIF neuronal models.

        Returns
        -------
        list
            Meta data for all GLIF neuronal models.
        '''

        include = "specimen(ephys_result[failed$eqfalse]),neuronal_model_template[name$il'*LIF*']"

        return self.model_query('NeuronalModel',
                                include=include,
                                num_rows='all')

    @deprecated()
    def get_neuronal_model(self, neuronal_model_id):
        '''DEPRECATED Query the current RMA endpoint with a neuronal_model id
        to get the corresponding well known files and meta data.

        Returns
        -------
        dict
            A dictionary containing
        '''


        include = ('neuronal_model_template(well_known_files(well_known_file_type)),' +
                   'specimen(ephys_sweeps,ephys_result(well_known_files(well_known_file_type))),' +
                   'well_known_files(well_known_file_type)')

        criteria = "[id$eq%d]" % neuronal_model_id

        self.neuronal_model = self.model_query('NeuronalModel',
                                               criteria=criteria,
                                               include=include,
                                               num_rows='all')[0]

        self.ephys_sweeps = None
        self.neuron_config_url = None
        self.stimulus_url = None

        # sweeps come from the specimen
        try:
            specimen = self.neuronal_model['specimen']
            self.ephys_sweeps = specimen['ephys_sweeps']
        except Exception as e:
            logging.info(e.args)
            self.ephys_sweeps = None

        if self.ephys_sweeps is None:
            logging.warning(
                "Could not find ephys_sweeps for this model (%d)" % self.neuronal_model['id'])

        # neuron config file comes from the neuronal model's well known files
        try:
            for wkf in self.neuronal_model['well_known_files']:
                if wkf['path'].endswith('neuron_config.json'):
                    self.neuron_config_url = wkf['download_link']
                    break
        except Exception as e:
            self.neuron_config_url = None

        if self.neuron_config_url is None:
            logging.warning(
                "Could not find neuron config well_known_file for this model (%d)" % self.neuronal_model['id'])

        # NWB file comes from the ephys_result's well known files
        try:
            ephys_result = specimen['ephys_result']
            for wkf in ephys_result['well_known_files']:
                if wkf['well_known_file_type']['name'] == 'NWBDownload':
                    self.stimulus_url = wkf['download_link']
                    break
        except Exception as e:
            self.stimulus_url = None

        if self.stimulus_url is None:
            logging.warning(
                "Could not find stimulus well_known_file for this model (%d)" % self.neuronal_model['id'])

        self.metadata = {
            'neuron_config_url': self.neuron_config_url,
            'stimulus_url': self.stimulus_url,
            'ephys_sweeps': self.ephys_sweeps,
            'neuronal_model': self.neuronal_model
        }

        return self.metadata

    @deprecated()
    def get_ephys_sweeps(self):
        ''' DEPRECATED Retrieve ephys sweep information out of downloaded metadata for a neuronal model

        Returns
        -------
        list
            A list of sweeps metadata dictionaries
        '''

        return self.ephys_sweeps

    @deprecated()
    def get_neuron_config(self, output_file_name=None):
        ''' DEPRECATED Retrieve a model configuration file from the API, optionally save it to disk, and
        return the contents of that file as a dictionary.

        Parameters
        ----------
        output_file_name: string
            File name to store the neuron configuration (optional).
        '''

        if self.neuron_config_url is None:
            raise Exception("URL for neuron config file is empty.")

        logging.info(self.api_url + self.neuron_config_url)

        neuron_config = self.retrieve_parsed_json_over_http(
            self.api_url + self.neuron_config_url)

        if output_file_name:
            with open(output_file_name, 'wb') as f:
                f.write(json.dumps(neuron_config, indent=2))

        return neuron_config

    @deprecated()
    def cache_stimulus_file(self, output_file_name):
        ''' DEPRECATED Download the NWB file for the current neuronal model and save it to a file.

        Parameters
        ----------
        output_file_name: string
            File name to store the NWB file.
        '''

        if self.stimulus_url is None:
            raise Exception("URL for stimulus file is empty.")

        self.retrieve_file_over_http(
            self.api_url + self.metadata['stimulus_url'], output_file_name)



# Allen Institute Software License - This software license is the 2-clause BSD
# license plus a third clause that prohibits redistribution for commercial
# purposes without further permission.
#
# Copyright 2017. Allen Institute. All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice,
# this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
# this list of conditions and the following disclaimer in the documentation
# and/or other materials provided with the distribution.
#
# 3. Redistributions for commercial purposes are not permitted without the
# Allen Institute's written permission.
# For purposes of this license, commercial purposes is the incorporation of the
# Allen Institute's software into anything for which you will charge fees or
# other compensation. Contact terms@alleninstitute.org for commercial licensing
# opportunities.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
#
from ..api import Api


class SvgApi(Api):

    def __init__(self, base_uri=None):
        super(SvgApi, self).__init__(base_uri)

    def build_query(self, section_image_id, groups=None, download=False):
        '''Build the URL that will fetch meta data for the specified structure.

        Parameters
        ----------
        section_image_id : integer
            Key of the object to be retrieved.
        groups : array of integers
            Keys of the group labels to filter the svg types that are returned.

        Returns
        -------
        url : string
            The constructed URL
        '''
        if download is True:
            endpoint = self.svg_download_endpoint
        else:
            endpoint = self.svg_endpoint

        if groups is None:
            groups = []

        if groups and len(groups) > 0:
            url_params = '?groups=' + ','.join([str(g) for g in groups])
        else:
            url_params = ''

        url = ''.join([endpoint,
                       '/',
                       str(section_image_id),
                       url_params])

        return url

    def download_svg(self,
                     section_image_id,
                     groups=None,
                     file_path=None):
        '''Download the svg file'''
        if file_path is None:
            file_path = '%d.svg' % (section_image_id)

        svg_url = self.build_query(section_image_id, groups, download=True)
        self.retrieve_file_over_http(svg_url, file_path)

    def get_svg(self,
                section_image_id,
                groups=None):
        '''Get the svg document.'''
        svg_url = self.build_query(section_image_id, groups)

        return self.retrieve_xml_over_http(svg_url)



# Allen Institute Software License - This software license is the 2-clause BSD
# license plus a third clause that prohibits redistribution for commercial
# purposes without further permission.
#
# Copyright 2014-2016. Allen Institute. All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice,
# this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
# this list of conditions and the following disclaimer in the documentation
# and/or other materials provided with the distribution.
#
# 3. Redistributions for commercial purposes are not permitted without the
# Allen Institute's written permission.
# For purposes of this license, commercial purposes is the incorporation of the
# Allen Institute's software into anything for which you will charge fees or
# other compensation. Contact terms@alleninstitute.org for commercial licensing
# opportunities.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
#
import logging
from allensdk.config.model.description import Description


class DescriptionParser(object):
    log = logging.getLogger(__name__)

    def __init__(self):
        pass

    def read(self, file_path, description=None, section=None, **kwargs):
        '''Parse data needed for a simulation.

        Parameters
        ----------
        description : dict
            Configuration from parsing previous files.
        section : string, optional
            What configuration section to read it into if the file does not specify.
        '''
        if description is None:
            description = Description()

        self.reader = self.parser_for_extension(file_path)
        self.reader.read(file_path, description, section, **kwargs)

        return description

    def read_string(self, data_string, description=None, section=None, header=None):
        '''Parse data needed for a simulation from a string.'''
        raise Exception("Not implemented, use a sub class")

    def write(self, filename, description):
        """Save the configuration.

        Parameters
        ----------
        filename : string
            Name of the file to write.
        """
        writer = self.parser_for_extension(filename)

        writer.write(filename, description)

    def parser_for_extension(self, filename):
        '''Choose a subclass that can read the format.

        Parameters
        ----------
        filename : string
           For the extension.

        Returns
        -------
        DescriptionParser
            Appropriate subclass.
        '''
        # Circular imports
        from allensdk.config.model.formats.json_description_parser import JsonDescriptionParser
        from allensdk.config.model.formats.pycfg_description_parser import PycfgDescriptionParser

        parser = None

        if filename.endswith('.json'):
            parser = JsonDescriptionParser()
        elif filename.endswith('.pycfg'):
            parser = PycfgDescriptionParser()
        else:
            raise Exception('could not determine file format')

        return parser



# Allen Institute Software License - This software license is the 2-clause BSD
# license plus a third clause that prohibits redistribution for commercial
# purposes without further permission.
#
# Copyright 2015-2017. Allen Institute. All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice,
# this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
# this list of conditions and the following disclaimer in the documentation
# and/or other materials provided with the distribution.
#
# 3. Redistributions for commercial purposes are not permitted without the
# Allen Institute's written permission.
# For purposes of this license, commercial purposes is the incorporation of the
# Allen Institute's software into anything for which you will charge fees or
# other compensation. Contact terms@alleninstitute.org for commercial licensing
# opportunities.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
#
from allensdk.config.manifest import Manifest, ManifestVersionError
import allensdk.core.json_utilities as ju
import pandas as pd
import pandas.io.json as pj
import functools
import os
import logging
from allensdk.deprecated import deprecated
import csv

from functools import wraps

def memoize(f):
   memodict = dict()

   @wraps(f)
   def wrapper(*args, **kwargs):
       key = (args, tuple(kwargs.items()))

       if key not in memodict:
           memodict[key] = f(*args, **kwargs)

       return memodict[key]

   return wrapper

class Cache(object):
    _log = logging.getLogger('allensdk.api.cache')

    def __init__(self,
                 manifest=None,
                 cache=True,
                 version=None):
        self.cache = cache
        self.load_manifest(manifest, version)

    def get_cache_path(self, file_name, manifest_key, *args):
        '''Helper method for accessing path specs from manifest keys.

        Parameters
        ----------
        file_name : string
        manifest_key : string
        args : ordered parameters

        Returns
        -------
        string or None
            path
        '''
        if self.cache:
            if file_name:
                return file_name
            elif self.manifest:
                return self.manifest.get_path(manifest_key, *args)

        return None

    def load_manifest(self, file_name, version=None):
        '''Read a keyed collection of path specifications.

        Parameters
        ----------
        file_name : string
            path to the manifest file

        Returns
        -------
        Manifest
        '''
        if file_name is not None:
            if not os.path.exists(file_name):

                # make the directory if it doesn't exist already
                dirname = os.path.dirname(file_name)
                if dirname:
                    Manifest.safe_mkdir(dirname)

                self.build_manifest(file_name)

            try:
                self.manifest = Manifest(
                    ju.read(file_name)['manifest'], 
                    os.path.dirname(file_name), 
                    version=version)
            except ManifestVersionError as e:
                raise ManifestVersionError(("Your manifest file (%s) is out of date" +
                                            " (version '%s' vs '%s').  Please remove this file" +
                                            " and it will be regenerated for you the next"
                                            " time you instantiate this class.") % (file_name, e.found_version, e.version),
                                           e.version, e.found_version)
        else:
            self.manifest = None

    def build_manifest(self, file_name):
        '''Creation of default path speifications.

        Parameters
        ----------
        file_name : string
            where to save it
        '''
        raise Exception(
            "This function must be defined in the appropriate subclass")

    def manifest_dataframe(self):
        '''Convenience method to view manifest as a pandas dataframe.
        '''
        return pd.DataFrame.from_dict(self.manifest.path_info,
                                      orient='index')

    @staticmethod
    def json_remove_keys(data, keys):
        for r in data:
            for key in keys:
                del r[key]

        return data

    @staticmethod
    def remove_keys(data, keys=None):
        ''' DataFrame version
        '''
        if keys is None:
            keys = []
        
        for key in keys:
            del data[key]

    @staticmethod
    def json_rename_columns(data,
                            new_old_name_tuples=None):
        '''Convenience method to rename columns in a pandas dataframe.

        Parameters
        ----------
        data : dataframe
            edited in place.
        new_old_name_tuples : list of string tuples (new, old)
        '''
        if new_old_name_tuples is None:
            new_old_name_tuples = []

        for new_name, old_name in new_old_name_tuples:
            for r in data:
                r[new_name] = r[old_name]
                del r[old_name]

    @staticmethod
    def rename_columns(data,
                       new_old_name_tuples=None):
        '''Convenience method to rename columns in a pandas dataframe.

        Parameters
        ----------
        data : dataframe
            edited in place.
        new_old_name_tuples : list of string tuples (new, old)
        '''
        if new_old_name_tuples is None:
            new_old_name_tuples = []

        for new_name, old_name in new_old_name_tuples:
            data.columns = [new_name if c == old_name else c
                            for c in data.columns]

    def load_csv(self,
                 path,
                 rename=None,
                 index=None):
        '''Read a csv file as a pandas dataframe.

        Parameters
        ----------
        rename : list of string tuples (new old), optional
            columns to rename
        index : string, optional
            post-rename column to use as the row label.
        '''
        data = pd.DataFrame.from_csv(path)

        Cache.rename_columns(data, rename)

        if index is not None:
            data.set_index([index], inplace=True)

        return data

    def load_json(self,
                  path,
                  rename=None,
                  index=None):
        '''Read a json file as a pandas dataframe.

        Parameters
        ----------
        rename : list of string tuples (new old), optional
            columns to rename
        index : string, optional
            post-rename column to use as the row label.
        '''
        data = pj.read_json(path, orient='records')

        Cache.rename_columns(data, rename)

        if index is not None:
            data.set_index([index], inplace=True)

        return data

    @staticmethod
    def cacher(fn,
               *args,
               **kwargs):
        '''make an rma query, save it and return the dataframe.
    
        Parameters
        ----------
        fn : function reference
            makes the actual query using kwargs.
        path : string
            where to save the data
        strategy : string or None, optional
            'create' always generates the data,
            'file' loads from disk,
            'lazy' queries the server if no file exists,
            None generates the data and bypasses all caching behavior
        pre : function
            df|json->df|json, takes one data argument and returns filtered version, None for pass-through
        post : function
            df|json->?, takes one data argument and returns Object
        reader : function, optional
            path -> data, default NOP
        writer : function, optional
            path, data -> None, default NOP
        kwargs : objects
            passed through to the query function

        Returns
        -------
        Object or None
            data type depends on fn, reader and/or post methods.
        '''
        path = kwargs.pop('path', None)
        strategy = kwargs.pop('strategy', None)
        pre = kwargs.pop('pre', lambda d: d)
        post = kwargs.pop('post', None)
        reader = kwargs.pop('reader', None)
        writer = kwargs.pop('writer', None)

        if strategy is None:
            if writer or path:
                strategy = 'lazy'
            else:
                strategy = 'pass_through'

        if not strategy in ['lazy', 'pass_through', 'file', 'create']:
            raise ValueError("Unknown query strategy: {}.".format(strategy))

        if 'lazy' == strategy:
            if os.path.exists(path):
                strategy = 'file'
            else:
                strategy = 'create'

        if strategy == 'pass_through':
                data = fn(*args, **kwargs)
        elif strategy in ['create']:
            Manifest.safe_make_parent_dirs(path)

            if writer:
                data = fn(*args, **kwargs)
                data = pre(data)
                writer(path, data)
            else:
                data = fn(*args, **kwargs)

        if reader:
            data = reader(path)

        # Note: don't provide post if fn or reader doesn't return data
        if post:
            data = post(data)
            return data

        try:
            data
            return data
        except:
            pass

        return

    @staticmethod
    def csv_writer(pth, gen):
        csv_writer = None
    
        first_row = True
        row_count = 1
    
        with open(pth, 'w') as output:
            for row in gen:
                if first_row:
                    field_names = map(str, row.keys())
                    csv_writer = csv.DictWriter(output,
                                                fieldnames=field_names,
                                                delimiter=',',
                                                quoting=csv.QUOTE_ALL)
                    csv_writer.writeheader()
                    first_row = False
                Cache._log.info('row: {}'.format(row_count))
                row_count = row_count + 1
                csv_writer.writerow(row)

    @staticmethod
    def cache_csv_json():
        return {
             'writer': Cache.csv_writer,
             'reader': lambda f: pd.DataFrame.from_csv(f).to_dict('records')
        }

    @staticmethod
    def cache_csv_dataframe():
        return {
             'writer': Cache.csv_writer,
             'reader' : pd.DataFrame.from_csv
        }

    @staticmethod
    def nocache_dataframe():
        return {
             'post': pd.DataFrame
        }

    @staticmethod
    def nocache_json():
        return {
        }

    @staticmethod
    def cache_json_dataframe():
        return {
             'writer': ju.write,
             'reader': lambda p: pj.read_json(p, orient='records')
        }

    @staticmethod
    def cache_json():
        return {
            'writer': ju.write,
            'reader' : ju.read
        }

    @staticmethod
    def cache_csv():
        return {
            'writer': Cache.csv_writer,
            'reader': pd.DataFrame.from_csv
        }

    @staticmethod
    def pathfinder(file_name_position,
                   secondary_file_name_position=None,
                   path_keyword=None):
        '''helper method to find path argument in legacy methods written
        prior to the @cacheable decorator.  Do not use for new @cacheable methods.
        
        Parameters
        ----------
        file_name_position : integer
            zero indexed position in the decorated method args where file path may be found.
        secondary_file_name_position : integer
            zero indexed position in the decorated method args where tha file path may be found.
        path_keyword : string
            kwarg that may have the file path.
        
        Notes
        -----
        This method is only intended to provide backward-compatibility for some
        methods that otherwise do not follow the path conventions of the @cacheable
        decorator.
        '''
        def pf(*args, **kwargs):
            file_name = None

            if path_keyword is not None and path_keyword in kwargs:
                file_name = kwargs[path_keyword]
            else:
                if file_name_position < len(args):
                    file_name = args[file_name_position]

                if (file_name is None and
                    secondary_file_name_position and 
                    secondary_file_name_position < len(args)):
                    file_name = args[secondary_file_name_position]

            return file_name
        return pf

    @deprecated()
    def wrap(self, fn, path, cache,
             save_as_json=True,
             return_dataframe=False,
             index=None,
             rename=None,
             **kwargs):
        '''make an rma query, save it and return the dataframe.
    
        Parameters
        ----------
        fn : function reference
            makes the actual query using kwargs.
        path : string
            where to save the data
        cache : boolean
            True will make the query, False just loads from disk
        save_as_json : boolean, optional
            True (default) will save data as json, False as csv
        return_dataframe : boolean, optional
            True will cast the return value to a pandas dataframe, False (default) will not
        index : string, optional
            column to use as the pandas index
        rename : list of string tuples, optional
            (new, old) columns to rename
        kwargs : objects
            passed through to the query function
    
        Returns
        -------
        dict or DataFrame
            data type depends on return_dataframe option.
    
        Notes
        -----
        Column renaming happens after the file is reloaded for json
        '''
        if cache is True:
            json_data = fn(**kwargs)

            if save_as_json is True:
                ju.write(path, json_data)
            else:
                df = pd.DataFrame(json_data)
                Cache.rename_columns(df, rename)

                if index is not None:
                    df.set_index([index], inplace=True)

                df.to_csv(path)

        # read it back in
        if save_as_json is True:
            if return_dataframe is True:
                data = pj.read_json(path, orient='records')
                Cache.rename_columns(data, rename)
                if index is not None:
                    data.set_index([index], inplace=True)
            else:
                data = ju.read(path)
        elif return_dataframe is True:
            data = pd.DataFrame.from_csv(path)
        else:
            raise ValueError(
                'save_as_json=False cannot be used with return_dataframe=False')
    
        return data


def cacheable(strategy=None,
              pre=None,
              writer=None,
              reader=None,
              post=None,
              pathfinder=None):
    '''decorator for rma queries, save it and return the dataframe.

    Parameters
    ----------
    fn : function reference
        makes the actual query using kwargs.
    path : string
        where to save the data
    strategy : string or None, optional
        'create' always gets the data from the source (server or generated),
        'file' loads from disk,
        'lazy' creates the data and saves to file if no file exists,
        None queries the server and bypasses all caching behavior
    pre : function
        df|json->df|json, takes one data argument and returns filtered version, None for pass-through
    post : function
        df|json->?, takes one data argument and returns Object
    reader : function, optional
        path -> data, default NOP
    writer : function, optional
        path, data -> None, default NOP
    kwargs : objects
        passed through to the query function

    Returns
    -------
    dict or DataFrame
        data type depends on dataframe option.

    Notes
    -----
    Column renaming happens after the file is reloaded for json
    '''
    def decor(func):
        decor.strategy=strategy
        decor.pre = pre
        decor.writer = writer
        decor.reader = reader
        decor.post = post
        decor.pathfinder = pathfinder

        @functools.wraps(func)
        def w(*args,
              **kwargs):
            if decor.pathfinder and not 'pathfinder' in kwargs:
                pathfinder = decor.pathfinder
            else:
                pathfinder = kwargs.pop('pathfinder', None)

            if pathfinder and not 'path' in kwargs:
                found_path = pathfinder(*args, **kwargs)
                
                if found_path:
                    kwargs['path'] = found_path
            if decor.strategy and not 'strategy' in kwargs:
                kwargs['strategy'] = decor.strategy
            if decor.pre and not 'pre' in kwargs:
                kwargs['pre'] = decor.pre
            if decor.writer and not 'writer' in kwargs:
                kwargs['writer'] = decor.writer
            if decor.reader and not 'reader' in kwargs:
                kwargs['reader'] = decor.reader
            if decor.post and not 'post in kwargs':
                kwargs['post'] = decor.post

            result = Cache.cacher(func,
                                  *args,
                                  **kwargs)
            return result
        
        return w
    return decor



# Allen Institute Software License - This software license is the 2-clause BSD
# license plus a third clause that prohibits redistribution for commercial
# purposes without further permission.
#
# Copyright 2017. Allen Institute. All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice,
# this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
# this list of conditions and the following disclaimer in the documentation
# and/or other materials provided with the distribution.
#
# 3. Redistributions for commercial purposes are not permitted without the
# Allen Institute's written permission.
# For purposes of this license, commercial purposes is the incorporation of the
# Allen Institute's software into anything for which you will charge fees or
# other compensation. Contact terms@alleninstitute.org for commercial licensing
# opportunities.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
#
import pytest
import mock
import numpy as np

from allensdk.core.reference_space import ReferenceSpace
from allensdk.core.structure_tree import StructureTree


@pytest.fixture
def rsp():

    tree = [{'id': 1, 'structure_id_path': [1]}, 
            {'id': 2, 'structure_id_path': [1, 2]}, 
            {'id': 3, 'structure_id_path': [1, 3]}, 
            {'id': 4, 'structure_id_path': [1, 2, 4]}, 
            {'id': 5, 'structure_id_path': [1, 2, 5]}, 
            {'id': 6, 'structure_id_path': [1, 2, 5, 6]}, 
            {'id': 7, 'structure_id_path': [1, 7]}]
            
    # leaves are 6, 4, 3
    # additionally annotate 2, 5 for realism :)
    annotation = np.zeros((10, 10, 10))
    annotation[4:8, 4:8, 4:8] = 2
    annotation[5:7, 5:7, 5:7] = 5
    annotation[5:7, 5:7, 5] = 6
    annotation[7, 7, 7] = 4
    annotation[8:10, 8:10, 8:10] = 3

    return ReferenceSpace(StructureTree(tree), annotation, [10, 10, 10])
    
    
def test_direct_voxel_counts(rsp):

    obt_one = rsp.direct_voxel_map
    obt_two = rsp.direct_voxel_map
    
    assert( obt_one[3] == 8 )
    assert( obt_one[2] == 4**3 - 2**3 - 1 )
    assert( obt_two[1] == 0 )
    assert( obt_two[2] == 4**3 - 2**3 - 1 )    

    
def test_total_voxel_counts(rsp):

    obt = rsp.total_voxel_map
    
    assert( obt[2] == 4**3 )
    assert( obt[6] == 4 )   
    
    
def test_remove_unassigned(rsp):

    rsp.remove_unassigned()
    node_ids = rsp.structure_tree.node_ids()
    
    assert( 1 in node_ids )
    assert( 7 not in node_ids )
    
    
def test_make_structure_mask(rsp):

    exp = np.zeros((10, 10, 10))
    exp[4:8, 4:8, 4:8] = 1
    exp[8:10, 8:10, 8:10] = 1
    obt = rsp.make_structure_mask([2, 3, 7])

    assert( np.allclose(obt, exp) )
    
    
def test_make_structure_mask_direct(rsp):

    exp = np.zeros((10, 10, 10))
    exp[5:7, 5:7, 6:7] = 1
    obt = rsp.make_structure_mask([5], True)

    assert( np.allclose(obt, exp) )
    
    
def test_many_structure_masks(rsp):

    cb = mock.MagicMock()
    
    [ii for ii in rsp.many_structure_masks([2, 3], output_cb=cb)]
    
    assert( cb.call_count == 2 )
    
    
def test_many_structure_masks_default_cb(rsp):
    
    rsp.make_structure_mask = mock.MagicMock(return_value=2)
    for item in rsp.many_structure_masks([1]):
        assert( np.allclose(item, [1, 2]) )
    
    
def test_check_coverage(rsp):
    
    mask = np.zeros((10, 10, 10))
    mask[7:10, 7:10, 7:10] = 1
    
    obt = rsp.check_coverage([3], mask)
    assert( np.count_nonzero(obt) == 27 - 8 )
    
    
def test_validate_structures(rsp):

    rsp.structure_tree.has_overlaps = mock.MagicMock()
    rsp.check_coverage = mock.MagicMock()
    
    rsp.validate_structures(1, 2)
    
    rsp.structure_tree.has_overlaps.assert_called_with(1)
    rsp.check_coverage.assert_called_with(1, 2)
    

def test_downsample(rsp):
    
    target = rsp.downsample((10, 20, 20))
    
    assert( np.allclose(target.annotation.shape, [10, 5, 5]) )
    
    
def test_get_slice_image(rsp):

    cmap = {0: [0, 0, 0], 1: [0, 0, 0], 2: [0, 0, 0], 3: [1, 2, 3], 
            4: [0, 0, 0], 5: [0, 0, 0], 6: [0, 0, 0], 7: [0, 0, 0], }
            
    image = rsp.get_slice_image(0, 90, cmap=cmap)
    
    assert( image[:, :, 0].sum() == 4 ) 
    
    
def test_direct_voxel_map_setter(rsp):
    
    rsp.direct_voxel_map = 4
    assert( rsp.direct_voxel_map == 4 )
    
    
def test_total_voxel_map_setter(rsp):
    
    rsp.total_voxel_map = 3
    assert( rsp.total_voxel_map == 3 )  



# Allen Institute Software License - This software license is the 2-clause BSD
# license plus a third clause that prohibits redistribution for commercial
# purposes without further permission.
#
# Copyright 2017. Allen Institute. All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice,
# this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
# this list of conditions and the following disclaimer in the documentation
# and/or other materials provided with the distribution.
#
# 3. Redistributions for commercial purposes are not permitted without the
# Allen Institute's written permission.
# For purposes of this license, commercial purposes is the incorporation of the
# Allen Institute's software into anything for which you will charge fees or
# other compensation. Contact terms@alleninstitute.org for commercial licensing
# opportunities.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
#
import allensdk.brain_observatory.circle_plots as cplots
import numpy as np

def test_polar_to_xy():
    d = cplots.polar_to_xy([0], 0.0)
    assert d.shape[0] == 1
    assert d.shape[1] == 2

    d = cplots.polar_to_xy([0, np.pi], 1.0)

    assert np.allclose(d, [[ 1, 0 ], [-1, 0]])

def test_polar_linspace():
    d = cplots.polar_linspace(1, 0, 180, 2, endpoint=True, degrees=True)
    assert np.allclose(d, [[1,0],[-1,0]])

    d = cplots.polar_linspace(1, 0, np.pi, 2, endpoint=False, degrees=False)
    assert np.allclose(d, [[1,0],[0,1.0]])

    d = cplots.polar_linspace(2, 0, 2*np.pi, 4, endpoint=False, degrees=False)
    assert np.allclose(d, [[2,0],[0,2],[-2,0],[0,-2]])

    d = cplots.polar_linspace(3, 0, 360, 5, endpoint=True, degrees=True)
    assert np.allclose(d, [[3,0],[0,3],[-3,0],[0,-3],[3,0]])

def test_spiral_trials():
    coll = cplots.spiral_trials([0,2])

    assert len(coll.get_paths()) == 2

def test_spiral_trials_polar():
    coll = cplots.spiral_trials_polar(1.0, 0.0, [1.0])
    assert len(coll.get_paths()) == 1

    coll = cplots.spiral_trials_polar(1.0, 0.0, [1.0], offset=[1,0])
    assert len(coll.get_paths()) == 1

def test_angle_lines():
    lines = cplots.angle_lines([0], 0, 1)
    assert len(lines.get_paths()) == 1

    lines = cplots.angle_lines([0,1], 0, 1)
    assert len(lines.get_paths()) == 2

def test_radial_arcs():
    arcs = cplots.radial_arcs([1], 0, 1)
    assert len(arcs.get_paths()) == 1

    arcs = cplots.radial_arcs([1,2], 0, 1)
    assert len(arcs.get_paths()) == 2

def test_radial_circles():
    d = cplots.radial_circles([1])
    assert len(d.get_paths()) == 1

    d = cplots.radial_circles([1,2])
    assert len(d.get_paths()) == 2

def test_polar_line_circles():
    d = cplots.polar_line_circles([1],0)
    assert len(d.get_paths()) == 1

    d = cplots.polar_line_circles([1],0,0)
    assert len(d.get_paths()) == 1

def test_wedge_ring():
    d = cplots.wedge_ring(1, 0, 1, 0, 180)
    assert len(d.get_paths()) == 1

    d = cplots.wedge_ring(2, 0, 1)
    assert len(d.get_paths()) == 2

def test_reset_hex_pack():
    pos = cplots.hex_pack(1.0, 1)
    cplots.reset_hex_pack()
    assert len(cplots.HEX_POSITIONS) == 0
    
def test_hex_pack():
    cplots.reset_hex_pack()

    pos = cplots.hex_pack(1.0, 1)
    assert pos.shape[0] == 1
    assert np.allclose(pos, [[0,0]])
    assert np.allclose(cplots.HEX_POSITIONS.shape, [1,2])

    pos = cplots.hex_pack(2.0, 2)
    assert np.allclose(pos, [[0,0],[4,0]])
    assert np.allclose(cplots.HEX_POSITIONS.shape, [7,2])

    pos = cplots.hex_pack(2.0, 8)
    assert np.allclose(cplots.HEX_POSITIONS.shape, [19,2])







# Allen Institute Software License - This software license is the 2-clause BSD
# license plus a third clause that prohibits redistribution for commercial
# purposes without further permission.
#
# Copyright 2015-2016. Allen Institute. All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice,
# this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
# this list of conditions and the following disclaimer in the documentation
# and/or other materials provided with the distribution.
#
# 3. Redistributions for commercial purposes are not permitted without the
# Allen Institute's written permission.
# For purposes of this license, commercial purposes is the incorporation of the
# Allen Institute's software into anything for which you will charge fees or
# other compensation. Contact terms@alleninstitute.org for commercial licensing
# opportunities.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
#
from .rma_api import RmaApi
from ..cache import cacheable
from allensdk.config.manifest import Manifest
from allensdk.api.cache import Cache


class CellTypesApi(RmaApi):
    NWB_FILE_TYPE = 'NWBDownload'
    SWC_FILE_TYPE = '3DNeuronReconstruction'
    MARKER_FILE_TYPE = '3DNeuronMarker'

    def __init__(self, base_uri=None):
        super(CellTypesApi, self).__init__(base_uri)

    def list_cells(self, 
                   id=None, 
                   require_morphology=False, 
                   require_reconstruction=False, 
                   reporter_status=None):
        """
        Query the API for a list of all cells in the Cell Types Database.

        Parameters
        ----------
        id: int
            ID of a cell.  If not provided returns all matching cells.  

        require_morphology: boolean
            Only return cells that have morphology images.

        require_reconstruction: boolean
            Only return cells that have morphological reconstructions.

        reporter_status: list
            Return cells that have a particular cell reporter status.

        Returns
        -------
        list
            Meta data for all cells.
        """

        if id:
            criteria = "[id$eq'%d']" % id
        else:
            criteria = "[is_cell_specimen$eq'true'],products[name$eq'Mouse Cell Types'],ephys_result[failed$eqfalse]"
        
        include = ('structure,donor(transgenic_lines),specimen_tags,cell_soma_locations,' +
                   'ephys_features,data_sets,neuron_reconstructions,cell_reporter')

        cells = self.model_query(
            'Specimen', criteria=criteria, include=include, num_rows='all')

        for cell in cells:
            # specimen tags
            for tag in cell['specimen_tags']:
                tag_name, tag_value = tag['name'].split(' - ')
                tag_name = tag_name.replace(' ', '_')
                cell[tag_name] = tag_value

            # morphology and reconstuction
            cell['has_reconstruction'] = len(
                cell['neuron_reconstructions']) > 0
            cell['has_morphology'] = len(cell['data_sets']) > 0

            # transgenic line
            cell['transgenic_line'] = None
            for tl in cell['donor']['transgenic_lines']:
                if tl['transgenic_line_type_name'] == 'driver':
                    cell['transgenic_line'] = tl['name']

            # cell reporter status
            cell['reporter_status'] = cell['cell_reporter']['name']

        result = self.filter_cells(cells, require_morphology, require_reconstruction, reporter_status)
        return result

    def get_cell(self, id):
        '''
        Query the API for a one cells in the Cell Types Database.

        
        Returns
        -------
        list
            Meta data for one cell.
        '''

        cells = self.list_cells(id=id)
        cell = None if not cells else cells[0]
        return cell

    @cacheable()
    def get_ephys_sweeps(self, specimen_id):
        """
        Query the API for a list of sweeps for a particular cell in the Cell Types Database.

        Parameters
        ----------
        specimen_id: int
            Specimen ID of a cell.

        Returns
        -------
        list: List of sweep dictionaries belonging to a cell
        """
        criteria = "[specimen_id$eq%d]" % specimen_id
        sweeps = self.model_query(
            'EphysSweep', criteria=criteria, num_rows='all')
        return sorted(sweeps, key=lambda x: x['sweep_number'])

    def filter_cells(self, cells, require_morphology, require_reconstruction, reporter_status):
        """
        Filter a list of cell specimens to those that optionally have morphologies
        or have morphological reconstructions.

        Parameters
        ----------

        cells: list
            List of cell metadata dictionaries to be filtered

        require_morphology: boolean
            Filter out cells that have no morphological images.

        require_reconstruction: boolean
            Filter out cells that have no morphological reconstructions.

        reporter_status: list
            Filter for cells that have a particular cell reporter status
        """

        if require_morphology:
            cells = [c for c in cells if c['has_morphology']]

        if require_reconstruction:
            cells = [c for c in cells if c['has_reconstruction']]

        if reporter_status:
            cells = [c for c in cells if c[
                'reporter_status'] in reporter_status]

        return cells

    @cacheable()
    def get_ephys_features(self):
        """
        Query the API for the full table of EphysFeatures for all cells.
        """

        return self.model_query(
            'EphysFeature',
            criteria='specimen(ephys_result[failed$eqfalse])',
            num_rows='all')

    @cacheable()
    def get_morphology_features(self):
        """
        Query the API for the full table of morphology features for all cells
        
        Notes
        -----
        by default the tags column is removed because it isn't useful
        """
        return self.model_query(
            'NeuronReconstruction',
            criteria="specimen(ephys_result[failed$eqfalse])",
            excpt='tags',
            num_rows='all')

    @cacheable(strategy='create',
               pathfinder=Cache.pathfinder(file_name_position=2,
                                           path_keyword='file_name'))
    def save_ephys_data(self, specimen_id, file_name):
        """
        Save the electrophysology recordings for a cell as an NWB file.

        Parameters
        ----------
        specimen_id: int
            ID of the specimen, from the Specimens database model in the Allen Institute API.

        file_name: str
            Path to save the NWB file.
        """
        criteria = '[id$eq%d],ephys_result(well_known_files(well_known_file_type[name$eq%s]))' % (
            specimen_id, self.NWB_FILE_TYPE)
        includes = 'ephys_result(well_known_files(well_known_file_type))'

        results = self.model_query('Specimen',
                                   criteria=criteria,
                                   include=includes,
                                   num_rows='all')

        try:
            file_url = results[0]['ephys_result'][
                'well_known_files'][0]['download_link']
        except Exception as _:
            raise Exception("Specimen %d has no ephys data" % specimen_id)

        self.retrieve_file_over_http(self.api_url + file_url, file_name)

    def save_reconstruction(self, specimen_id, file_name):
        """
        Save the morphological reconstruction of a cell as an SWC file.

        Parameters
        ----------
        specimen_id: int
            ID of the specimen, from the Specimens database model in the Allen Institute API.

        file_name: str
            Path to save the SWC file.
        """

        Manifest.safe_make_parent_dirs(file_name)

        criteria = '[id$eq%d],neuron_reconstructions(well_known_files)' % specimen_id
        includes = 'neuron_reconstructions(well_known_files(well_known_file_type[name$eq\'%s\']))' % self.SWC_FILE_TYPE

        results = self.model_query('Specimen',
                                   criteria=criteria,
                                   include=includes,
                                   num_rows='all')

        try:
            file_url = results[0]['neuron_reconstructions'][
                0]['well_known_files'][0]['download_link']
        except:
            raise Exception("Specimen %d has no reconstruction" % specimen_id)

        self.retrieve_file_over_http(self.api_url + file_url, file_name)

    def save_reconstruction_markers(self, specimen_id, file_name):
        """
        Save the marker file for the morphological reconstruction of a cell.  These are
        comma-delimited files indicating points of interest in a reconstruction (truncation
        points, early tracing termination, etc).

        Parameters
        ----------
        specimen_id: int
            ID of the specimen, from the Specimens database model in the Allen Institute API.

        file_name: str
            Path to save the marker file.
        """

        Manifest.safe_make_parent_dirs(file_name)

        criteria = '[id$eq%d],neuron_reconstructions(well_known_files)' % specimen_id
        includes = 'neuron_reconstructions(well_known_files(well_known_file_type[name$eq\'%s\']))' % self.MARKER_FILE_TYPE

        results = self.model_query('Specimen',
                                   criteria=criteria,
                                   include=includes,
                                   num_rows='all')

        try:
            file_url = results[0]['neuron_reconstructions'][
                0]['well_known_files'][0]['download_link']
        except:
            raise LookupError("Specimen %d has no marker file" % specimen_id)

        self.retrieve_file_over_http(self.api_url + file_url, file_name)



# Allen Institute Software License - This software license is the 2-clause BSD
# license plus a third clause that prohibits redistribution for commercial
# purposes without further permission.
#
# Copyright 2017. Allen Institute. All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice,
# this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
# this list of conditions and the following disclaimer in the documentation
# and/or other materials provided with the distribution.
#
# 3. Redistributions for commercial purposes are not permitted without the
# Allen Institute's written permission.
# For purposes of this license, commercial purposes is the incorporation of the
# Allen Institute's software into anything for which you will charge fees or
# other compensation. Contact terms@alleninstitute.org for commercial licensing
# opportunities.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
#
import pytest
import os
import matplotlib.image as mpimg
import numpy as np
from allensdk.core.brain_observatory_nwb_data_set import BrainObservatoryNwbDataSet
import allensdk.brain_observatory.observatory_plots as oplots
from allensdk.brain_observatory.static_gratings import StaticGratings
from allensdk.brain_observatory.drifting_gratings import DriftingGratings
from allensdk.brain_observatory.natural_scenes import NaturalScenes
from allensdk.brain_observatory.natural_movie import NaturalMovie
from allensdk.brain_observatory.locally_sparse_noise import LocallySparseNoise
import allensdk.brain_observatory.stimulus_info as stiminfo
import allensdk.core.json_utilities as ju
from pkg_resources import resource_filename  # @UnresolvedImport

if 'TEST_OBSERVATORY_EXPERIMENT_PLOTS_DATA' in os.environ:
    data_file = os.environ['TEST_OBSERVATORY_EXPERIMENT_PLOTS_DATA']
else:
    data_file = resource_filename(__name__, 'test_observatory_plots_data.json')

if data_file == 'skip':
    EXPERIMENT_CONTAINER=None
    TEST_DATA_DIR=None
else:
    EXPERIMENT_CONTAINER = ju.read(data_file)
    TEST_DATA_DIR = EXPERIMENT_CONTAINER['image_directory']

class AnalysisSingleton(object):
    def __init__(self, klass, session, *args):
        self.klass = klass
        self.session = session
        self.args = args

        self.obj = None

    @staticmethod
    def experiment_for_session(session):
        return next(exp for exp in EXPERIMENT_CONTAINER['experiments'] if exp['session'] == session)

    def __call__(self):
        if self.obj is None:
            exp = self.experiment_for_session(self.session)
            data_set = BrainObservatoryNwbDataSet(exp['nwb_file'])
            self.obj = self.klass.from_analysis_file(data_set, exp['analysis_file'], *self.args)

        return self.obj

STATIC_GRATINGS = AnalysisSingleton(StaticGratings, stiminfo.THREE_SESSION_B)
DRIFTING_GRATINGS = AnalysisSingleton(DriftingGratings, stiminfo.THREE_SESSION_A)
NATURAL_SCENES = AnalysisSingleton(NaturalScenes, stiminfo.THREE_SESSION_B)
NATURAL_MOVIE_ONE_A = AnalysisSingleton(NaturalMovie, stiminfo.THREE_SESSION_A, stiminfo.NATURAL_MOVIE_ONE)
NATURAL_MOVIE_ONE_B = AnalysisSingleton(NaturalMovie, stiminfo.THREE_SESSION_B, stiminfo.NATURAL_MOVIE_ONE)
NATURAL_MOVIE_ONE_C = AnalysisSingleton(NaturalMovie, stiminfo.THREE_SESSION_C, stiminfo.NATURAL_MOVIE_ONE)
NATURAL_MOVIE_TWO = AnalysisSingleton(NaturalMovie, stiminfo.THREE_SESSION_C, stiminfo.NATURAL_MOVIE_TWO)
NATURAL_MOVIE_THREE = AnalysisSingleton(NaturalMovie, stiminfo.THREE_SESSION_A, stiminfo.NATURAL_MOVIE_THREE)
LOCALLY_SPARSE_NOISE = AnalysisSingleton(LocallySparseNoise, stiminfo.THREE_SESSION_C, stiminfo.LOCALLY_SPARSE_NOISE)

if EXPERIMENT_CONTAINER:
    CELL_SPECIMEN_ID = EXPERIMENT_CONTAINER['cells'][0]
else:
    CELL_SPECIMEN_ID = None


def assert_images_match(new_file, test_file, shape):
    assert os.path.exists(new_file)
    new_img = mpimg.imread(new_file)
    assert np.allclose(new_img.shape[:2], shape)
    
    assert os.path.exists(test_file)
    test_img = mpimg.imread(test_file)
    assert np.allclose(new_img.shape, test_img.shape)
    assert np.allclose(test_img.shape[:2], shape)

    assert (new_img - test_img).mean() < 0.1
    
    os.remove(new_file)


@pytest.mark.skipif(data_file == 'skip', reason='NWB Data files not configured')
@pytest.mark.parametrize("new_file,static_gratings", 
                         [ [ 'static_gratings_ttp.png', STATIC_GRATINGS ] ])
def test_ttp_static_gratings(new_file, static_gratings, shape=[500,500]):
    with oplots.figure_in_px(shape[1], shape[0], new_file) as fig:
        static_gratings().plot_time_to_peak()
        oplots.finalize_with_axes()
    assert_images_match(new_file, os.path.join(TEST_DATA_DIR, new_file), shape)


@pytest.mark.skipif(data_file == 'skip', reason='NWB Data files not configured')
@pytest.mark.parametrize("new_file,static_gratings", 
                         [ [ 'static_gratings_pref_ori.png', STATIC_GRATINGS ] ])
def test_pref_ori_static_gratings(new_file, static_gratings, shape=[250,500]):
    with oplots.figure_in_px(shape[1], shape[0], new_file) as fig:
        static_gratings().plot_preferred_orientation()
        oplots.finalize_no_axes()
    assert_images_match(new_file, os.path.join(TEST_DATA_DIR, new_file), shape)


@pytest.mark.skipif(data_file == 'skip', reason='NWB Data files not configured')
@pytest.mark.parametrize("new_file,static_gratings", 
                         [ [ 'static_gratings_ori.png', STATIC_GRATINGS ] ])
def test_osi_static_gratings(new_file, static_gratings, shape=[500,500]):
    with oplots.figure_in_px(shape[1], shape[0], new_file) as fig:
        static_gratings().plot_orientation_selectivity()
        oplots.finalize_with_axes()
    assert_images_match(new_file, os.path.join(TEST_DATA_DIR, new_file), shape)


@pytest.mark.skipif(data_file == 'skip', reason='NWB Data files not configured')
@pytest.mark.parametrize("new_file,static_gratings", 
                         [ [ 'static_gratings_pref_sf.png', STATIC_GRATINGS ] ])
def test_pref_sf(new_file, static_gratings, shape=[500,500]):
    with oplots.figure_in_px(shape[1], shape[0], new_file) as fig:
        static_gratings().plot_preferred_spatial_frequency()
        oplots.finalize_with_axes()
    assert_images_match(new_file, os.path.join(TEST_DATA_DIR, new_file), shape)


@pytest.mark.skipif(data_file == 'skip', reason='NWB Data files not configured')
@pytest.mark.parametrize("new_file,drifting_gratings", 
                         [ [ 'drifting_gratings_pref_dir.png', DRIFTING_GRATINGS ] ])
def test_pref_dir_drifting_gratings(new_file, drifting_gratings, shape=[500,500]):
    with oplots.figure_in_px(shape[1], shape[0], new_file) as fig:
        drifting_gratings().plot_preferred_direction()
        oplots.finalize_no_axes()
    assert_images_match(new_file, os.path.join(TEST_DATA_DIR, new_file), shape)


@pytest.mark.skipif(data_file == 'skip', reason='NWB Data files not configured')
@pytest.mark.parametrize("new_file,drifting_gratings", 
                         [ [ 'drifting_gratings_pref_tf.png', DRIFTING_GRATINGS ] ])
def test_pref_tf_drifting_gratings(new_file, drifting_gratings, shape=[500,500]):
    with oplots.figure_in_px(shape[1], shape[0], new_file) as fig:
        drifting_gratings().plot_preferred_temporal_frequency()
        oplots.finalize_with_axes()
    assert_images_match(new_file, os.path.join(TEST_DATA_DIR, new_file), shape)


@pytest.mark.skipif(data_file == 'skip', reason='NWB Data files not configured')
@pytest.mark.parametrize("new_file,drifting_gratings", 
                         [ [ 'drifting_gratings_dsi.png', DRIFTING_GRATINGS ] ])
def test_dsi_drifting_gratings(new_file, drifting_gratings, shape=[500,500]):
    with oplots.figure_in_px(shape[1], shape[0], new_file) as fig:
        drifting_gratings().plot_direction_selectivity()
        oplots.finalize_with_axes()
    assert_images_match(new_file, os.path.join(TEST_DATA_DIR, new_file), shape)


@pytest.mark.skipif(data_file == 'skip', reason='NWB Data files not configured')
@pytest.mark.parametrize("new_file,drifting_gratings", 
                         [ [ 'drifting_gratings_osi.png', DRIFTING_GRATINGS ] ])
def test_osi_drifting_gratings(new_file, drifting_gratings, shape=[500,500]):
    with oplots.figure_in_px(shape[1], shape[0], new_file) as fig:
        drifting_gratings().plot_orientation_selectivity()
        oplots.finalize_with_axes()
    assert_images_match(new_file, os.path.join(TEST_DATA_DIR, new_file), shape)


@pytest.mark.skipif(data_file == 'skip', reason='NWB Data files not configured')
@pytest.mark.parametrize("new_file,natural_scenes", 
                         [ [ 'natural_scenes_ttp.png', NATURAL_SCENES ] ])
def test_ttp_natural_scenes(new_file, natural_scenes, shape=[500,500]):
    with oplots.figure_in_px(shape[1], shape[0], new_file) as fig:
        natural_scenes().plot_time_to_peak()
        oplots.finalize_with_axes()
    assert_images_match(new_file, os.path.join(TEST_DATA_DIR, new_file), shape)


@pytest.mark.skipif(data_file == 'skip', reason='NWB Data files not configured')
@pytest.mark.parametrize("new_file,static_gratings,cell_specimen_id",
                         [ [ 'static_gratings_fan_plot.png', STATIC_GRATINGS, CELL_SPECIMEN_ID ] ])
def test_fan_plot(new_file, static_gratings, cell_specimen_id, shape=[250,500]):
    with oplots.figure_in_px(shape[1], shape[0], new_file) as fig:
        static_gratings().open_fan_plot(cell_specimen_id)
        oplots.finalize_no_axes()
    assert_images_match(new_file, os.path.join(TEST_DATA_DIR, new_file), shape)


@pytest.mark.skipif(data_file == 'skip', reason='NWB Data files not configured')
@pytest.mark.parametrize("new_file,natural_scenes,cell_specimen_id", 
                         [ [ 'natural_scenes_fan_plot.png', NATURAL_SCENES, CELL_SPECIMEN_ID ] ])
def test_corona_plot(new_file, natural_scenes, cell_specimen_id, shape=[500,500]):
    with oplots.figure_in_px(shape[1], shape[0], new_file) as fig:
        natural_scenes().open_corona_plot(cell_specimen_id)
        oplots.finalize_no_axes()
    assert_images_match(new_file, os.path.join(TEST_DATA_DIR, new_file), shape)


@pytest.mark.skipif(data_file == 'skip', reason='NWB Data files not configured')
@pytest.mark.parametrize("new_file,natural_movie,cell_specimen_id", 
                         [ ("natural_movie_one_a_track_plot.png", NATURAL_MOVIE_ONE_A, CELL_SPECIMEN_ID),
                           ("natural_movie_one_b_track_plot.png", NATURAL_MOVIE_ONE_B, CELL_SPECIMEN_ID),
                           ("natural_movie_one_c_track_plot.png", NATURAL_MOVIE_ONE_C, CELL_SPECIMEN_ID),
                           ("natural_movie_two_track_plot.png", NATURAL_MOVIE_TWO, CELL_SPECIMEN_ID),
                           ("natural_movie_three_track_plot.png", NATURAL_MOVIE_THREE, CELL_SPECIMEN_ID) ])
def test_track_plot(new_file, natural_movie, cell_specimen_id, shape=[500,500]):
    with oplots.figure_in_px(shape[1], shape[0], new_file) as fig:
        natural_movie().open_track_plot(cell_specimen_id)
        oplots.finalize_no_axes()
    assert_images_match(new_file, os.path.join(TEST_DATA_DIR, new_file), shape)


@pytest.mark.skipif(data_file == 'skip', reason='NWB Data files not configured')
@pytest.mark.parametrize("new_file,drifting_gratings,cell_specimen_id", 
                         [ [ 'drifting_gratings_star_plot.png', DRIFTING_GRATINGS, CELL_SPECIMEN_ID ] ])
def test_star_plot(new_file, drifting_gratings, cell_specimen_id, shape=[500,500]):
    with oplots.figure_in_px(shape[1], shape[0], new_file) as fig:
        drifting_gratings().open_star_plot(cell_specimen_id)
        oplots.finalize_no_axes()
    assert_images_match(new_file, os.path.join(TEST_DATA_DIR, new_file), shape)


@pytest.mark.skipif(data_file == 'skip', reason='NWB Data files not configured')
@pytest.mark.parametrize("new_file,analysis,cell_specimen_id", 
                         [ ("3sa_speed_tuning_plot.png", DRIFTING_GRATINGS, CELL_SPECIMEN_ID),
                           ("3sb_speed_tuning_plot.png", STATIC_GRATINGS, CELL_SPECIMEN_ID),
                           ("3sc_speed_tuning_plot.png", NATURAL_MOVIE_TWO, CELL_SPECIMEN_ID) ])
def test_speed_tuning_plot(new_file, analysis, cell_specimen_id, shape=[500,500]):
    with oplots.figure_in_px(shape[1], shape[0], new_file) as fig:
        analysis().plot_speed_tuning(cell_specimen_id)
        oplots.finalize_with_axes()
    assert_images_match(new_file, os.path.join(TEST_DATA_DIR, new_file), shape)


@pytest.mark.skipif(data_file == 'skip', reason='NWB Data files not configured')
@pytest.mark.parametrize("new_file,locally_sparse_noise,on,cell_specimen_id", 
                         [ ('locally_sparse_noise_on.png', LOCALLY_SPARSE_NOISE, True, CELL_SPECIMEN_ID),
                           ('locally_sparse_noise_off.png', LOCALLY_SPARSE_NOISE, False, CELL_SPECIMEN_ID) ])
def test_pincushion_plot(new_file, locally_sparse_noise, on, cell_specimen_id, shape=[500,877]):
    with oplots.figure_in_px(shape[1], shape[0], new_file) as fig:
        locally_sparse_noise().open_pincushion_plot(on, cell_specimen_id)
        oplots.finalize_no_axes()
    assert_images_match(new_file, os.path.join(TEST_DATA_DIR, new_file), shape)



# Allen Institute Software License - This software license is the 2-clause BSD
# license plus a third clause that prohibits redistribution for commercial
# purposes without further permission.
#
# Copyright 2017. Allen Institute. All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice,
# this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
# this list of conditions and the following disclaimer in the documentation
# and/or other materials provided with the distribution.
#
# 3. Redistributions for commercial purposes are not permitted without the
# Allen Institute's written permission.
# For purposes of this license, commercial purposes is the incorporation of the
# Allen Institute's software into anything for which you will charge fees or
# other compensation. Contact terms@alleninstitute.org for commercial licensing
# opportunities.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
#
import pytest
from mock import MagicMock
import allensdk.core.json_utilities as ju
from allensdk.api.queries.rma_template import RmaTemplate


@pytest.fixture
def rma():
    templates = \
        {"ontology_queries": [
            {'name': 'structures_by_graph_ids',
             'description': 'see name',
             'model': 'Structure',
             'criteria': '[graph_id$in{{ graph_ids }}]',
             'order': ['structures.graph_order'],
             'num_rows': 'all',
             'count': False,
             'criteria_params': ['graph_ids']
             },
            {'name': 'structures_by_graph_names',
             'description': 'see name',
             'model': 'Structure',
             'criteria': 'graph[structure_graphs.name$in{{ graph_names }}]',
             'order': ['structures.graph_order'],
             'num_rows': 'all',
             'count': False,
             'criteria_params': ['graph_names']
             },
            {'name': 'structures_by_set_ids',
             'description': 'see name',
             'model': 'Structure',
             'criteria': '[structure_set_id$in{{ set_ids }}]',
             'order': ['structures.graph_order'],
             'num_rows': 'all',
             'count': False,
             'criteria_params': ['set_ids']
             },
            {'name': 'structures_by_set_names',
             'description': 'see name',
             'model': 'Structure',
             'criteria': 'structure_sets[name$in{{ set_names }}]',
             'order': ['structures.graph_order'],
             'num_rows': 'all',
             'count': False,
             'criteria_params': ['set_names']
             },
            {'name': 'structure_graphs_list',
             'description': 'see name',
             'model': 'StructureGraph',
             'num_rows': 'all',
             'count': False
             },
            {'name': 'structure_sets_list',
             'description': 'see name',
             'model': 'StructureSet',
             'num_rows': 'all',
             'count': False
             },
            {'name': 'atlases_list',
             'description': 'see name',
             'model': 'Atlas',
             'num_rows': 'all',
             'count': False
             },
            {'name': 'atlases_table',
             'description': 'see name',
             'model': 'Atlas',
             'criteria': '{% if graph_ids is defined %}[graph_id$in{{ graph_ids }}],{% endif %}structure_graph(ontology),graphic_group_labels',
             'include': '[structure_graph(ontology),graphic_group_labels',
             'num_rows': 'all',
             'count': False,
             'criteria_params': ['graph_ids']
             },
            {'name': 'atlases_table_brief',
             'description': 'see name',
             'model': 'Atlas',
             'criteria': 'structure_graph(ontology),graphic_group_labels',
             'include': 'structure_graph(ontology),graphic_group_labels',
             'only': ['atlases.id',
                      'atlases.name',
                      'atlases.image_type',
                      'ontologies.id',
                      'ontologies.name',
                      'structure_graphs.id',
                      'structure_graphs.name',
                      'graphic_group_labels.id',
                      'graphic_group_labels.name'],
             'num_rows': 'all',
             'count': False
             }
        ]}
    rma = RmaTemplate(query_manifest=templates)

    return rma


def test_atlases_list(rma):
    ju.read_url_get = \
        MagicMock(name='read_url_get',
                  return_value={'msg': [{'whatever': True}]})

    rma.template_query('ontology_queries',
                       'atlases_list')

    ju.read_url_get.assert_called_once_with(
        "http://api.brain-map.org/api/v2/data/query.json?q="
        "model::Atlas,rma::options"
        "%5Bnum_rows$eq%27all%27%5D%5Bcount$eqfalse%5D")


def test_structure_graphs_list(rma):
    ju.read_url_get = \
        MagicMock(name='read_url_get',
                  return_value={'msg': [{'whatever': True}]})

    rma.template_query('ontology_queries',
                       'structure_graphs_list')

    ju.read_url_get.assert_called_once_with(
        "http://api.brain-map.org/api/v2/data/query.json?q="
        "model::StructureGraph,rma::options"
        "%5Bnum_rows$eq%27all%27%5D%5Bcount$eqfalse%5D")


def test_structure_sets_list(rma):
    ju.read_url_get = \
        MagicMock(name='read_url_get',
                  return_value={'msg': [{'whatever': True}]})

    rma.template_query('ontology_queries',
                       'structure_sets_list')

    ju.read_url_get.assert_called_once_with(
        "http://api.brain-map.org/api/v2/data/query.json?q="
        "model::StructureSet,rma::options"
        "%5Bnum_rows$eq%27all%27%5D%5Bcount$eqfalse%5D")


def test_structures_by_graph_ids(rma):
    ju.read_url_get = \
        MagicMock(name='read_url_get',
                  return_value={'msg': [{'whatever': True}]})

    rma.template_query('ontology_queries',
                       'structures_by_graph_ids',
                       graph_ids='1')

    ju.read_url_get.assert_called_once_with(
        "http://api.brain-map.org/api/v2/data/query.json?q="
        "model::Structure,rma::criteria,"
        "%5Bgraph_id$in1%5D,rma::options"
        "%5Bnum_rows$eq%27all%27%5D%5Border$eqstructures.graph_order%5D"
        "%5Bcount$eqfalse%5D")


def test_structures_by_two_graph_ids(rma):
    ju.read_url_get = \
        MagicMock(name='read_url_get',
                  return_value={'msg': [{'whatever': True}]})

    rma.template_query('ontology_queries',
                       'structures_by_graph_ids',
                       graph_ids=[1, 2])

    ju.read_url_get.assert_called_once_with(
        "http://api.brain-map.org/api/v2/data/query.json?q="
        "model::Structure,rma::criteria,"
        "%5Bgraph_id$in1,2%5D,"
        "rma::options"
        "%5Bnum_rows$eq%27all%27%5D"
        "%5Border$eqstructures.graph_order%5D%5Bcount$eqfalse%5D")


def test_structures_by_graph_names(rma):
    ju.read_url_get = \
        MagicMock(name='read_url_get',
                  return_value={'msg': [{'whatever': True}]})

    rma.template_query('ontology_queries',
                       'structures_by_graph_names',
                       graph_names=rma.quote_string('Human+Brain+Atlas'))

    ju.read_url_get.assert_called_once_with(
        "http://api.brain-map.org/api/v2/data/query.json?q="
        "model::Structure,rma::criteria,"
        "graph%5Bstructure_graphs.name$in%27Human+Brain+Atlas%27%5D,"
        "rma::options"
        "%5Bnum_rows$eq%27all%27%5D"
        "%5Border$eqstructures.graph_order%5D%5Bcount$eqfalse%5D")


def test_structures_by_set_ids(rma):
    ju.read_url_get = \
        MagicMock(name='read_url_get',
                  return_value={'msg': [{'whatever': True}]})

    rma.template_query('ontology_queries',
                       'structures_by_graph_ids',
                       graph_ids='1')

    ju.read_url_get.assert_called_once_with(
        "http://api.brain-map.org/api/v2/data/query.json?q="
        "model::Structure,rma::criteria,"
        "%5Bgraph_id$in1%5D,rma::options%5Bnum_rows$eq%27all%27%5D"
        "%5Border$eqstructures.graph_order%5D%5Bcount$eqfalse%5D")


def test_atlases_table(rma):
    ju.read_url_get = \
        MagicMock(name='read_url_get',
                  return_value={'msg': [{'whatever': True}]})

    rma.template_query('ontology_queries',
                       'atlases_table')

    ju.read_url_get.assert_called_once_with(
        "http://api.brain-map.org/api/v2/data/query.json?q="
        "model::Atlas,rma::criteria,"
        "structure_graph%28ontology%29,graphic_group_labels,"
        "rma::include,%5Bstructure_graph%28ontology%29,graphic_group_labels,"
        "rma::options%5Bnum_rows$eq%27all%27%5D%5Bcount$eqfalse%5D")


def test_atlases_table_one_graph(rma):
    ju.read_url_get = \
        MagicMock(name='read_url_get',
                  return_value={'msg': [{'whatever': True}]})

    rma.template_query('ontology_queries',
                       'atlases_table',
                       graph_ids=1)

    ju.read_url_get.assert_called_once_with(
        "http://api.brain-map.org/api/v2/data/query.json?q="
        "model::Atlas,rma::criteria,"
        "%5Bgraph_id$in1%5D,structure_graph%28ontology%29,graphic_group_labels,"
        "rma::include,%5Bstructure_graph%28ontology%29,graphic_group_labels,"
        "rma::options%5Bnum_rows$eq%27all%27%5D%5Bcount$eqfalse%5D")


def test_atlases_table_brief(rma):
    ju.read_url_get = \
        MagicMock(name='read_url_get',
                  return_value={'msg': [{'whatever': True}]})

    rma.template_query('ontology_queries',
                       'atlases_table_brief')

    ju.read_url_get.assert_called_once_with(
        "http://api.brain-map.org/api/v2/data/query.json?q="
        "model::Atlas,"
        "rma::criteria,structure_graph%28ontology%29,graphic_group_labels,"
        "rma::include,structure_graph%28ontology%29,graphic_group_labels,"
        "rma::options%5Bonly$eq%27atlases.id,atlases.name,atlases.image_type,"
        "ontologies.id,ontologies.name,structure_graphs.id,structure_graphs.name,"
        "graphic_group_labels.id,graphic_group_labels.name%27%5D%5B"
        "num_rows$eq%27all%27%5D%5Bcount$eqfalse%5D")



# Allen Institute Software License - This software license is the 2-clause BSD
# license plus a third clause that prohibits redistribution for commercial
# purposes without further permission.
#
# Copyright 2017. Allen Institute. All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice,
# this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
# this list of conditions and the following disclaimer in the documentation
# and/or other materials provided with the distribution.
#
# 3. Redistributions for commercial purposes are not permitted without the
# Allen Institute's written permission.
# For purposes of this license, commercial purposes is the incorporation of the
# Allen Institute's software into anything for which you will charge fees or
# other compensation. Contact terms@alleninstitute.org for commercial licensing
# opportunities.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
#
import pytest
from mock import MagicMock
from allensdk.api.queries.grid_data_api import GridDataApi


@pytest.fixture
def grid_data():
    gda = GridDataApi()
    gda.retrieve_file_over_http = \
        MagicMock(name='retrieve_file_over_http')

    return gda


def test_api_doc_url_download_expression_grid(grid_data):
    '''Url to download the 200um density volume
       for the Mouse Brain Atlas SectionDataSet 69816930.

    Notes
    -----
    See `Downloading 3-D Expression Grid Data <http://help.brain-map.org/display/api/Downloading+3-D+Expression+Grid+Data#Downloading3-DExpressionGridData-DOWNLOADING3DEXPRESSIONGRIDDATA>`_
    , example 'Download the 200um density volume for the Mouse Brain Atlas SectionDataSet 69816930'.
    '''
    path = '69816930.zip'
    section_data_set_id = 69816930
    grid_data.download_expression_grid_data(section_data_set_id)
    expected = 'http://api.brain-map.org/grid_data/download/69816930'
    grid_data.retrieve_file_over_http.assert_called_once_with(expected, path)


def test_api_doc_url_download_expression_grid_energy_intensity(grid_data):
    '''Url to download the 200um energy and intensity volumes for Mouse Brain Atlas SectionDataSet 69816930.

    Notes
    -----
    See `Downloading 3-D Expression Grid Data <http://help.brain-map.org/display/api/Downloading+3-D+Expression+Grid+Data#Downloading3-DExpressionGridData-DOWNLOADING3DEXPRESSIONGRIDDATA>`_
    , example 'Download the 200um energy and intensity volumes for Mouse Brain Atlas SectionDataSet 69816930'.

    The id in the example url doesn't match the caption.
    '''
    path = '183282970.zip'
    section_data_set_id = 183282970
    include = ['energy', 'intensity']
    grid_data.download_expression_grid_data(section_data_set_id,
                                            include=include)

    grid_data.retrieve_file_over_http.assert_called_once_with(
        "http://api.brain-map.org/grid_data/download/183282970"
        "?include=energy,intensity",
        path)


def test_api_doc_url_projection_grid(grid_data):
    '''Url to download the 100um density volume for the Mouse Connectivity Atlas SectionDataSet 181777177.

    Notes
    -----
    See `Downloading 3-D Projection Grid Data <http://help.brain-map.org/display/api/Downloading+3-D+Expression+Grid+Data#Downloading3-DExpressionGridData-DOWNLOADING3DPROJECTIONGRIDDATA>`_
    , example 'Download the 100um density volume for the Mouse Connectivity Atlas SectionDataSet 181777177'.
    '''
    path = '181777177.nrrd'
    section_data_set_id = 181777177
    grid_data.download_projection_grid_data(section_data_set_id)
    expected = 'http://api.brain-map.org/grid_data/download_file/181777177'
    grid_data.retrieve_file_over_http.assert_called_once_with(expected, path)


def test_api_doc_url_projection_grid_injection_fraction_resolution(grid_data):
    '''Url to download the 25um injection_fraction volume for Mouse Connectivity Atlas SectionDataSet 181777177.

    Notes
    -----
    See `Downloading 3-D Projection Grid Data <http://help.brain-map.org/display/api/Downloading+3-D+Expression+Grid+Data#Downloading3-DExpressionGridData-DOWNLOADING3DPROJECTIONGRIDDATA>`_
    , example 'Download the 25um injection_fraction volume for Mouse Connectivity Atlas SectionDataSet 181777177'.
    '''
    section_data_set_id = 181777177
    path = 'id.nrrd'
    grid_data.download_projection_grid_data(section_data_set_id,
                                            [grid_data.INJECTION_FRACTION],
                                            resolution=25,
                                            save_file_path=path)

    grid_data.retrieve_file_over_http.assert_called_once_with(
        "http://api.brain-map.org/grid_data/download_file/181777177"
        "?image=injection_fraction&resolution=25",
        path)



# Allen Institute Software License - This software license is the 2-clause BSD
# license plus a third clause that prohibits redistribution for commercial
# purposes without further permission.
#
# Copyright 2015-2016. Allen Institute. All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice,
# this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
# this list of conditions and the following disclaimer in the documentation
# and/or other materials provided with the distribution.
#
# 3. Redistributions for commercial purposes are not permitted without the
# Allen Institute's written permission.
# For purposes of this license, commercial purposes is the incorporation of the
# Allen Institute's software into anything for which you will charge fees or
# other compensation. Contact terms@alleninstitute.org for commercial licensing
# opportunities.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
#
from ..api import Api


class SynchronizationApi(Api):
    '''HTTP client for image synchronization services uses the image alignment results from
    the Informatics Data Processing Pipeline.
    Note: all locations on SectionImages are reported in pixel coordinates
    and all locations in 3-D ReferenceSpaces are reported in microns.

    See `Image to Image Synchronization <http://help.brain-map.org/display/api/Image-to-Image+Synchronization>`_
    for additional documentation.
    '''

    def __init__(self, base_uri=None):
        super(SynchronizationApi, self).__init__(base_uri)

    def get_image_to_atlas(self,
                           section_image_id,
                           x, y,
                           atlas_id):
        '''For a specified Atlas, find the closest annotated SectionImage
        and (x,y) location as defined by a seed SectionImage and seed (x,y) location.

        Parameters
        ----------
        section_image_id : integer
            Seed for spatial sync.
        x : float
            Pixel coordinate of the seed location in the seed SectionImage.
        y : float
            Pixel coordinate of the seed location in the seed SectionImage.
        atlas_id : int
            Target Atlas for image sync.

        Returns
        -------
        dict
            The parsed json response
        '''
        url = ''.join([self.image_to_atlas_endpoint,
                       '/',
                       str(section_image_id),
                       '.json',
                       '?x=%f&y=%f' % (x, y),
                       '&atlas_id=',
                       str(atlas_id)])

        return self.json_msg_query(url)

    def get_image_to_image(self,
                           section_image_id,
                           x, y,
                           section_data_set_ids):
        '''For a list of target SectionDataSets, find the closest SectionImage
        and (x,y) location as defined by a seed SectionImage and seed (x,y) pixel location.

        Parameters
        ----------
        section_image_id : integer
            Seed for spatial sync.
        x : float
            Pixel coordinate of the seed location in the seed SectionImage.
        y : float
            Pixel coordinate of the seed location in the seed SectionImage.
        section_data_set_ids : list of integers
            Target SectionDataSet IDs for image sync.

        Returns
        -------
        dict
            The parsed json response
        '''
        url = ''.join([self.image_to_image_endpoint,
                       '/',
                       str(section_image_id),
                       '.json',
                       '?x=%f&y=%f' % (x, y),
                       '&section_data_set_ids=',
                       ','.join(str(i) for i in section_data_set_ids)])

        return self.json_msg_query(url)

    def get_image_to_image_2d(self,
                              section_image_id,
                              x, y,
                              section_image_ids):
        '''For a list of target SectionImages, find the closest (x,y) location
        as defined by a seed SectionImage and seed (x,y) location.

        Parameters
        ----------
        section_image_id : integer
            Seed for image sync.
        x : float
            Pixel coordinate of the seed location in the seed SectionImage.
        y : float
            Pixel coordinate of the seed location in the seed SectionImage.
        section_image_ids : list of ints
            Target SectionImage IDs for image sync.

        Returns
        -------
        dict
            The parsed json response
        '''
        url = ''.join([self.image_to_image_2d_endpoint,
                       '/',
                       str(section_image_id),
                       '.json',
                       '?x=%f&y=%f' % (x, y),
                       '&section_image_ids=',
                       ','.join(str(i) for i in section_image_ids)])

        return self.json_msg_query(url)

    def get_reference_to_image(self,
                               reference_space_id,
                               x, y, z,
                               section_data_set_ids):
        '''For a list of target SectionDataSets, find the closest SectionImage
        and (x,y) location as defined by a (x,y,z) location in a specified ReferenceSpace.

        Parameters
        ----------
        reference_space_id : integer
            Seed for spatial sync.
        x : float
            Coordinate (in microns) of the seed location in the seed ReferenceSpace.
        y : float
            Coordinate (in microns) of the seed location in the seed ReferenceSpace.
        z : float
            Coordinate (in microns) of the seed location in the seed ReferenceSpace.
        section_data_set_ids : list of ints
            Target SectionDataSets IDs for image sync.

        Returns
        -------
        dict
            The parsed json response
        '''
        url = ''.join([self.reference_to_image_endpoint,
                       '/',
                       str(reference_space_id),
                       '.json',
                       '?x=%f&y=%f&z=%f' % (x, y, z),
                       '&section_data_set_ids=',
                       ','.join(str(i) for i in section_data_set_ids)])

        return self.json_msg_query(url)

    def get_image_to_reference(self,
                               section_image_id,
                               x, y):
        '''For a specified SectionImage and (x,y) location,
        return the (x,y,z) location in the ReferenceSpace of the associated SectionDataSet.

        Parameters
        ----------
        section_image_id : integer
            Seed for image sync.
        x : float
            Pixel coordinate on the specified SectionImage.
        y : float
            Pixel coordinate on the specified SectionImage.

        Returns
        -------
        dict
            The parsed json response
        '''
        url = ''.join([self.image_to_reference_endpoint,
                       '/',
                       str(section_image_id),
                       '.json',
                       '?x=%f&y=%f' % (x, y)])

        return self.json_msg_query(url)

    def get_structure_to_image(self,
                               section_data_set_id,
                               structure_ids):
        '''For a list of target structures, find the closest SectionImage
        and (x,y) location as defined by the centroid of each Structure.

        Parameters
        ----------
        section_data_set_id : integer
            primary key
        structure_ids : list of integers
            primary key

        Returns
        -------
        dict
            The parsed json response
        '''
        url = ''.join([self.structure_to_image_endpoint,
                       '/',
                       str(section_data_set_id),
                       '.json',
                       '?structure_ids=',
                       ','.join([str(i) for i in structure_ids])])

        return self.json_msg_query(url)



import pytest
import numpy as np
import os
from allensdk.core.brain_observatory_nwb_data_set import BrainObservatoryNwbDataSet, si
import numpy as np
from pkg_resources import resource_filename  # @UnresolvedImport
NWB_FLAVORS = []

if 'TEST_NWB_FILES' in os.environ:
    nwb_list_file = os.environ['TEST_NWB_FILES']
else:
    nwb_list_file = resource_filename(__name__, os.path.join('..','core','nwb_files.txt'))

if nwb_list_file == 'skip':
    NWB_FLAVORS = []
else:
    with open(nwb_list_file, 'r') as f:
        NWB_FLAVORS = [l.strip() for l in f]

@pytest.fixture(params=NWB_FLAVORS)
def data_set(request):
    data_set = BrainObservatoryNwbDataSet(request.param)

    return data_set

def test_BinaryIntervalSearchTree():

    bist = si.BinaryIntervalSearchTree([(0, .9, 'A'), (1, 1.9, 'B'), (3, 3.9, 'D'), (2, 2.9, 'C')])
    assert bist.search(1.5)[2] == 'B'
    assert bist.search(0)[2] == 'A'
    assert bist.search(2.5)[2] == 'C'
    assert bist.search(3.5)[2] == 'D'

def test_pixels_to_visual_degrees():
    m = si.BrainObservatoryMonitor()
    np.testing.assert_almost_equal(m.pixels_to_visual_degrees(1), 0.103270443661,10)

@pytest.mark.skipif(not os.path.exists('/projects/neuralcoding'),
                    reason="test NWB file not available")
def test_StimulusSearch(data_set):

    epoch_df = data_set.get_stimulus_epoch_table()
    s = si.StimulusSearch(data_set)
    assert len(s.search(epoch_df.iloc[2]['end'])) == 3
    assert s.search(epoch_df.iloc[2]['end'] + 1) is None
    assert len(s.search(752)) == 3


def test_sessions_with_stimulus():

    for session_type, stimulus_type_list in si.SESSION_STIMULUS_MAP.items():
        for stimulus_type in stimulus_type_list:
            assert session_type in si.sessions_with_stimulus(stimulus_type)

def test_stimuli_in_session():

    for session_type, stimulus_type_list in si.SESSION_STIMULUS_MAP.items():
        for stimulus_type in stimulus_type_list:
            assert session_type in si.sessions_with_stimulus(stimulus_type)

def test_stimuli_in_session():

    test_dict = {si.THREE_SESSION_A:4,
                 si.THREE_SESSION_B:4,
                 si.THREE_SESSION_C:4,
                 si.THREE_SESSION_C2:5}

    for session in si.SESSION_LIST:
        assert session in si.SESSION_STIMULUS_MAP
        assert len(si.stimuli_in_session(session)) == test_dict[session]
    assert len(si.SESSION_STIMULUS_MAP) == len(si.SESSION_LIST) == 4

def test_all_stimuli():
    assert len(si.all_stimuli()) == 10

def test_rotate():
    np.testing.assert_array_almost_equal(np.array(si.rotate(1,1,np.pi)), np.array([-1,-1]))

def test_get_spatial_grating():

    data = si.get_spatial_grating(height=100, aspect_ratio=2, ori=45, pix_per_cycle=10, phase=0, p2p_amp=2, baseline=1)

    assert data.shape == (100,200)
    np.testing.assert_almost_equal(data[0,0], data[-1,-1])
    np.testing.assert_almost_equal(data.max(), 2, 3)
    np.testing.assert_almost_equal(data.min(), 0, 3)
    np.testing.assert_almost_equal(data[50,100], 2)

def test_get_spatio_temporal_grating():

    for t, test_val in zip([0,.5,1], [2,0,2]):
        data = si.get_spatio_temporal_grating(t, height=100, aspect_ratio=2, ori=45, pix_per_cycle=10, phase=0, p2p_amp=2, baseline=1, temporal_frequency=1)
        np.testing.assert_almost_equal(data[50,100], test_val)

    data = si.get_spatio_temporal_grating(0, height=100,
                                             aspect_ratio=2,
                                             ori=45,
                                             pix_per_cycle=20,
                                             phase=0,
                                             p2p_amp=2,
                                             baseline=1,
                                             temporal_frequency=1)

    x1 = data[50, 100]

    data = si.get_spatio_temporal_grating(.5, height=100,
                                             aspect_ratio=2,
                                             ori=45,
                                             pix_per_cycle=20,
                                             phase=.5,
                                             p2p_amp=2,
                                             baseline=1,
                                             temporal_frequency=1)

    x2 = data[50, 100]

    np.testing.assert_almost_equal(x1, x2)

def test_map_template_monitor():



    np.testing.assert_almost_equal(np.array((500, 250)),
                                   si.map_template_coordinate_to_monitor_coordinate((20, 20), (1000, 500), (40, 40)))


    np.testing.assert_almost_equal(np.array((20,20)),
                                   si.map_monitor_coordinate_to_template_coordinate((500, 250), (1000, 500), (40,40)))

def test_lsn_monitor():

    lsn4_template_coordinate = (8, 14)
    lsn4_monitor_coordinate = np.array(si.MONITOR_DIMENSIONS)/2 #(600,960)
    np.testing.assert_almost_equal(np.array(lsn4_monitor_coordinate),
                                   si.map_stimulus_coordinate_to_monitor_coordinate(lsn4_template_coordinate, si.MONITOR_DIMENSIONS, si.LOCALLY_SPARSE_NOISE_4DEG))

    lsn4_template_coordinate = (4, 7)
    lsn4_monitor_coordinate = np.array(si.MONITOR_DIMENSIONS)/2#(600,960)
    np.testing.assert_almost_equal(np.array(lsn4_monitor_coordinate),
                                   si.map_stimulus_coordinate_to_monitor_coordinate(lsn4_template_coordinate, si.MONITOR_DIMENSIONS, si.LOCALLY_SPARSE_NOISE_8DEG))

    lsn4_template_coordinate = (0,0)
    lsn4_monitor_coordinate = (240,330)
    np.testing.assert_almost_equal(np.array(lsn4_monitor_coordinate),
                                   si.map_stimulus_coordinate_to_monitor_coordinate(lsn4_template_coordinate,
                                                                           si.MONITOR_DIMENSIONS,
                                                                           si.LOCALLY_SPARSE_NOISE_4DEG))

    lsn4_template_coordinate = (0,0)
    lsn4_monitor_coordinate = (240,330)
    np.testing.assert_almost_equal(np.array(lsn4_template_coordinate),
                                   si.monitor_coordinate_to_lsn_coordinate(lsn4_monitor_coordinate,
                                                                           si.MONITOR_DIMENSIONS,
                                                                           si.LOCALLY_SPARSE_NOISE_4DEG))


    lsn4_template_coordinate = (0,0)
    lsn4_monitor_coordinate = (240,330)
    np.testing.assert_almost_equal(np.array(lsn4_monitor_coordinate),
                                   si.map_stimulus_coordinate_to_monitor_coordinate(lsn4_template_coordinate,
                                                                           si.MONITOR_DIMENSIONS,
                                                                           si.LOCALLY_SPARSE_NOISE_8DEG))

    lsn4_template_coordinate = (0,0)
    lsn4_monitor_coordinate = (240,330)
    np.testing.assert_almost_equal(np.array(lsn4_template_coordinate),
                                   si.monitor_coordinate_to_lsn_coordinate(lsn4_monitor_coordinate,
                                                                           si.MONITOR_DIMENSIONS,
                                                                           si.LOCALLY_SPARSE_NOISE_8DEG))

def test_natural_scene_monitor():

    template_coordinate = (0,0)
    monitor_coordinate = (141, 373)
    np.testing.assert_almost_equal(np.array(monitor_coordinate),
                                   si.natural_scene_coordinate_to_monitor_coordinate(template_coordinate,
                                                                                     si.MONITOR_DIMENSIONS))

    template_coordinate = (0,0)
    monitor_coordinate = (141, 373)
    np.testing.assert_almost_equal(np.array(template_coordinate),
                                   si.map_monitor_coordinate_to_stimulus_coordinate(monitor_coordinate,
                                                                                    si.MONITOR_DIMENSIONS,
                                                                                    si.NATURAL_SCENES))

def test_natural_movie_monitor():

    template_coordinate = (0,0)
    monitor_coordinate = (60, 0)
    np.testing.assert_almost_equal(np.array(monitor_coordinate),
                                   si.natural_movie_coordinate_to_monitor_coordinate(template_coordinate,
                                                                                     si.MONITOR_DIMENSIONS))

    template_coordinate = (0,0)
    monitor_coordinate = (60, 0)
    np.testing.assert_almost_equal(np.array(template_coordinate),
                                   si.map_monitor_coordinate_to_stimulus_coordinate(monitor_coordinate,
                                                                                    si.MONITOR_DIMENSIONS,
                                                                                    si.NATURAL_MOVIE_ONE))

def test_bijective_all_stimuli():

    for stimulus in si.all_stimuli():

        template_coordinate = (10,10)
        monitor_coordinate = si.map_stimulus_coordinate_to_monitor_coordinate(template_coordinate,
                                                                              si.MONITOR_DIMENSIONS,
                                                                              stimulus)

        new_template_coordinate = si.map_monitor_coordinate_to_stimulus_coordinate(monitor_coordinate,
                                                                                   si.MONITOR_DIMENSIONS,
                                                                                   stimulus)

        np.testing.assert_array_almost_equal(template_coordinate, new_template_coordinate)

        for original_loc in [(0,0), (10,10)]:

            for target_stimulus in si.all_stimuli():

                new_loc = si.map_stimulus(original_loc, stimulus, target_stimulus, si.MONITOR_DIMENSIONS)
                new_original_loc = si.map_stimulus(new_loc, target_stimulus, stimulus, si.MONITOR_DIMENSIONS)
                np.testing.assert_array_almost_equal(new_original_loc, original_loc)

def test_monitor_basic_spatial_unit():

    m = si.Monitor(300,400, 5, 'cm')
    m.set_spatial_unit('cm')

    m.set_spatial_unit('inch')
    np.testing.assert_almost_equal(m.panel_size, 1.968505, 5)
    np.testing.assert_almost_equal(1./m.aspect_ratio, 3./4)
    np.testing.assert_almost_equal(m.height, 0.46500143220300011)
    np.testing.assert_almost_equal(m.width, 0.62000190960400015)
    np.testing.assert_almost_equal(m.pixel_size, 0.0015500047740100004)

    m.set_spatial_unit('cm')
    np.testing.assert_almost_equal(m.panel_size, 5)
    np.testing.assert_almost_equal(1./m.aspect_ratio, 3./4)
    np.testing.assert_almost_equal(m.height, 3)
    np.testing.assert_almost_equal(m.width, 4)
    np.testing.assert_almost_equal(m.pixel_size, .01)


def test_pixels_to_visual_degrees():

    m = si.BrainObservatoryMonitor()

    np.testing.assert_almost_equal(m.pixels_to_visual_degrees(45), 4.64716996476)
    np.testing.assert_almost_equal(m.pixels_to_visual_degrees(45, small_angle_approximation=False), 4.64462483116)

    np.testing.assert_almost_equal(m.pixels_to_visual_degrees(1), 0.103270443661)
    np.testing.assert_almost_equal(m.pixels_to_visual_degrees(1, small_angle_approximation=False), 0.103270415704)

@pytest.mark.skipif(not os.path.exists('/projects/neuralcoding'),
                    reason="test NWB file not available")
def test_lsn_image_to_screen(data_set):

    compare_set = set(data_set.list_stimuli()).intersection(si.LOCALLY_SPARSE_NOISE_STIMULUS_TYPES)
    if len(compare_set) > 0:
        for stimulus_type in compare_set:

            template = data_set.get_stimulus_template(stimulus_type)
            m = si.BrainObservatoryMonitor()
            m.lsn_image_to_screen(template[0,:,:]).shape == si.MONITOR_DIMENSIONS

@pytest.mark.skipif(not os.path.exists('/projects/neuralcoding'),
                    reason="test NWB file not available")
def test_natural_movie_image_to_screen(data_set):

    compare_set = set(data_set.list_stimuli()).intersection(si.NATURAL_MOVIE_STIMULUS_TYPES)
    if len(compare_set) > 0:
        for stimulus_type in compare_set:

            template = data_set.get_stimulus_template(stimulus_type)
            m = si.BrainObservatoryMonitor()
            m.natural_movie_image_to_screen(template[0, :, :]).shape == si.MONITOR_DIMENSIONS

@pytest.mark.skipif(not os.path.exists('/projects/neuralcoding'),
                    reason="test NWB file not available")
def test_grating_to_screen(data_set):

    compare_set = set(data_set.list_stimuli()).intersection([si.STATIC_GRATINGS, si.DRIFTING_GRATINGS])
    if len(compare_set) > 0:

        for stimulus_type in compare_set:
            m = si.BrainObservatoryMonitor()
            curr_row = data_set.get_stimulus_table(stimulus_type).iloc[10]
            phase = 0
            spatial_frequency = .04
            orientation = curr_row.orientation
            template =  m.grating_to_screen(phase, spatial_frequency, orientation)
            assert m.natural_movie_image_to_screen(template).shape == si.MONITOR_DIMENSIONS

def test_get_mask():
    m = si.BrainObservatoryMonitor()
    mask = m.get_mask()

    assert mask.sum() == 931286
    assert mask.shape == si.MONITOR_DIMENSIONS

def test_spatial_frequency_to_pix_per_cycle():

    m = si.BrainObservatoryMonitor()

    x1 = m.spatial_frequency_to_pix_per_cycle(.1, 15.0)
    x2 = m.spatial_frequency_to_pix_per_cycle(.05, 15.0)

    np.testing.assert_almost_equal(x1, 97.7072500845)
    np.testing.assert_almost_equal(x2/x1, 2)

def test_show_image():

    m = si.BrainObservatoryMonitor()

    img = np.zeros(si.MONITOR_DIMENSIONS)
    m.show_image(img, show=False, warp=True, mask=False)
    m.show_image(img, show=False, warp=False, mask=True)

def test_map_stimulus():

    m = si.BrainObservatoryMonitor()
    test_list = [(0, 0), (-5.333333333333333, -7.333333333333333), (-5.333333333333333, -7.333333333333333), (-2.6666666666666665, -3.6666666666666665), (-16.88888888888889, 0.0), (-16.88888888888889, 0.0), (-16.88888888888889, 0.0), (-141.0, -373.0), (0, 0), (0, 0), (240.0, 330.0), (0.0, 0.0), (0.0, 0.0), (0.0, 0.0), (50.666666666666664, 104.5), (50.666666666666664, 104.5), (50.666666666666664, 104.5), (99.0, -43.0), (240.0, 330.0), (240.0, 330.0), (240.0, 330.0), (0.0, 0.0), (0.0, 0.0), (0.0, 0.0), (50.666666666666664, 104.5), (50.666666666666664, 104.5), (50.666666666666664, 104.5), (99.0, -43.0), (240.0, 330.0), (240.0, 330.0), (240.0, 330.0), (0.0, 0.0), (0.0, 0.0), (0.0, 0.0), (50.666666666666664, 104.5), (50.666666666666664, 104.5), (50.666666666666664, 104.5), (99.0, -43.0), (240.0, 330.0), (240.0, 330.0), (60.0, 0.0), (-4.0, -7.333333333333333), (-4.0, -7.333333333333333), (-2.0, -3.6666666666666665), (0.0, 0.0), (0.0, 0.0), (0.0, 0.0), (-81.0, -373.0), (60.0, 0.0), (60.0, 0.0), (60.0, 0.0), (-4.0, -7.333333333333333), (-4.0, -7.333333333333333), (-2.0, -3.6666666666666665), (0.0, 0.0), (0.0, 0.0), (0.0, 0.0), (-81.0, -373.0), (60.0, 0.0), (60.0, 0.0), (60.0, 0.0), (-4.0, -7.333333333333333), (-4.0, -7.333333333333333), (-2.0, -3.6666666666666665), (0.0, 0.0), (0.0, 0.0), (0.0, 0.0), (-81.0, -373.0), (60.0, 0.0), (60.0, 0.0), (141.0, 373.0), (-2.2, 0.9555555555555556), (-2.2, 0.9555555555555556), (-1.1, 0.4777777777777778), (22.8, 118.11666666666666), (22.8, 118.11666666666666), (22.8, 118.11666666666666), (0.0, 0.0), (141.0, 373.0), (141.0, 373.0), (0, 0), (-5.333333333333333, -7.333333333333333), (-5.333333333333333, -7.333333333333333), (-2.6666666666666665, -3.6666666666666665), (-16.88888888888889, 0.0), (-16.88888888888889, 0.0), (-16.88888888888889, 0.0), (-141.0, -373.0), (0, 0), (0, 0), (0, 0), (-5.333333333333333, -7.333333333333333), (-5.333333333333333, -7.333333333333333), (-2.6666666666666665, -3.6666666666666665), (-16.88888888888889, 0.0), (-16.88888888888889, 0.0), (-16.88888888888889, 0.0), (-141.0, -373.0), (0, 0), (0, 0)]
    counter = 0
    for source_stimulus in sorted(si.all_stimuli()):
        for target_stimulus in sorted(si.all_stimuli()):
            tmp = m.map_stimulus((0,0), source_stimulus, target_stimulus)
            np.testing.assert_array_almost_equal(tmp, test_list[counter])
            counter += 1
            np.testing.assert_array_almost_equal(m.map_stimulus(tmp, target_stimulus, source_stimulus), np.array([0,0]))

if __name__ == "__main__":
#
    # with open(nwb_list_file, 'r') as f:
    #     NWB_FLAVORS = [l.strip() for l in f]
    #
    # for nwb_file_location in NWB_FLAVORS:
    #     data_set = BrainObservatoryNwbDataSet(nwb_file_location)
    #     test_lsn_image_to_screen(data_set)
    #     test_natural_movie_image_to_screen(data_set)
    #     test_grating_to_screen(data_set)

    # test_StimulusSearch()
    # test_BinaryIntervalSearchTree()
    # test_sessions_with_stimulus()
    # test_stimuli_in_session()
    # test_all_stimuli()
    # test_rotate()
    # test_get_spatial_grating()
    # test_get_spatio_temporal_grating()
    # test_map_template_coordinate_to_monitor_coordinate()
    # test_natural_scene_monitor()
    # test_bijective_all_stimuli()
    # test_monitor_basic_spatial_unit()
    # test_brain_observatory_monitor()
    # test_spatial_frequency_to_pix_per_cycle()
    # test_get_mask()
    # test_show_image()
    test_map_stimulus()


# Allen Institute Software License - This software license is the 2-clause BSD
# license plus a third clause that prohibits redistribution for commercial
# purposes without further permission.
#
# Copyright 2017. Allen Institute. All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice,
# this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
# this list of conditions and the following disclaimer in the documentation
# and/or other materials provided with the distribution.
#
# 3. Redistributions for commercial purposes are not permitted without the
# Allen Institute's written permission.
# For purposes of this license, commercial purposes is the incorporation of the
# Allen Institute's software into anything for which you will charge fees or
# other compensation. Contact terms@alleninstitute.org for commercial licensing
# opportunities.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
#
import numpy as np
from .static_gratings import StaticGratings
from .locally_sparse_noise import LocallySparseNoise
from .natural_scenes import NaturalScenes
from .drifting_gratings import DriftingGratings
from .natural_movie import NaturalMovie
import six
from allensdk.core.brain_observatory_nwb_data_set \
    import BrainObservatoryNwbDataSet
from . import stimulus_info
from allensdk.brain_observatory.brain_observatory_exceptions \
    import BrainObservatoryAnalysisException
from . import brain_observatory_plotting as cp
import argparse
import logging
import os



def multi_dataframe_merge(dfs):
    """ merge a number of pd.DataFrames into a single dataframe on their index columns. 
    If any columns are duplicated, prefer the first occuring instance of the column """

    out_df = None
    for _, df in enumerate(dfs):
        if out_df is None:
            out_df = df
        else:
            out_df = out_df.merge(df, left_index=True,
                                  right_index=True, suffixes=['', '_deleteme'])

    bad_columns = set([c for c in out_df.columns if c.endswith('deleteme')])
    out_df.drop(list(bad_columns), axis=1, inplace=True)

    return out_df


class SessionAnalysis(object):
    """ 
    Run all of the stimulus-specific analyses associated with a single experiment session. 

    Parameters
    ----------
    nwb_path: string, path to NWB file

    save_path: string, path to HDF5 file to store outputs.  Recommended NOT to modify the NWB file.
    """

    _log = logging.getLogger('allensdk.brain_observatory.session_analysis')

    def __init__(self, nwb_path, save_path):
        self.nwb = BrainObservatoryNwbDataSet(nwb_path)
        self.save_path = save_path
        self.save_dir = os.path.dirname(save_path)

        self.metrics_a = dict(cell={},experiment={})
        self.metrics_b = dict(cell={},experiment={})
        self.metrics_c = dict(cell={},experiment={})

        self.metadata = self.nwb.get_metadata()

    def append_metadata(self, df):
        """ Append the metadata fields from the NWB file as columns to a pd.DataFrame """

        for k, v in six.iteritems(self.metadata):
            df[k] = v

    def save_session_a(self, dg, nm1, nm3, peak):
        """ Save the output of session A analysis to self.save_path.  

        Parameters
        ----------
        dg: DriftingGratings instance

        nm1: NaturalMovie instance
            This NaturalMovie instance should have been created with
            movie_name = stimulus_info.NATURAL_MOVIE_ONE

        nm3: NaturalMovie instance
            This NaturalMovie instance should have been created with
            movie_name = stimulus_info.NATURAL_MOVIE_THREE

        peak: pd.DataFrame
            The combined peak response property table created in self.session_a().
        """

        nwb = BrainObservatoryNwbDataSet(self.save_path)

        nwb.save_analysis_dataframes(
            ('stim_table_dg', dg.stim_table),
            ('sweep_response_dg', dg.sweep_response),
            ('mean_sweep_response_dg', dg.mean_sweep_response),
            ('peak', peak),
            ('sweep_response_nm1', nm1.sweep_response),
            ('stim_table_nm1', nm1.stim_table),
            ('sweep_response_nm3', nm3.sweep_response))

        nwb.save_analysis_arrays(
            ('response_dg', dg.response),
            ('binned_cells_sp', nm1.binned_cells_sp),
            ('binned_cells_vis', nm1.binned_cells_vis),
            ('binned_dx_sp', nm1.binned_dx_sp),
            ('binned_dx_vis', nm1.binned_dx_vis),
            ('noise_corr_dg', dg.noise_correlation),
            ('signal_corr_dg', dg.signal_correlation),
            ('rep_similarity_dg', dg.representational_similarity)
            )


    def save_session_b(self, sg, nm1, ns, peak):
        """ Save the output of session B analysis to self.save_path.  

        Parameters
        ----------
        sg: StaticGratings instance

        nm1: NaturalMovie instance
            This NaturalMovie instance should have been created with
            movie_name = stimulus_info.NATURAL_MOVIE_ONE

        ns: NaturalScenes instance

        peak: pd.DataFrame
            The combined peak response property table created in self.session_b().
        """

        nwb = BrainObservatoryNwbDataSet(self.save_path)

        nwb.save_analysis_dataframes(
            ('stim_table_sg', sg.stim_table),
            ('sweep_response_sg', sg.sweep_response),
            ('mean_sweep_response_sg', sg.mean_sweep_response),
            ('sweep_response_nm1', nm1.sweep_response),
            ('stim_table_nm1', nm1.stim_table),
            ('sweep_response_ns', ns.sweep_response),
            ('stim_table_ns', ns.stim_table),
            ('mean_sweep_response_ns', ns.mean_sweep_response),
            ('peak', peak))

        nwb.save_analysis_arrays(
            ('response_sg', sg.response),
            ('response_ns', ns.response),
            ('binned_cells_sp', nm1.binned_cells_sp),
            ('binned_cells_vis', nm1.binned_cells_vis),
            ('binned_dx_sp', nm1.binned_dx_sp),
            ('binned_dx_vis', nm1.binned_dx_vis),
            ('noise_corr_sg', sg.noise_correlation),
            ('signal_corr_sg', sg.signal_correlation),
            ('rep_similarity_sg', sg.representational_similarity),
            ('noise_corr_ns', ns.noise_correlation),
            ('signal_corr_ns', ns.signal_correlation),
            ('rep_similarity_ns', ns.representational_similarity)
            )

    def save_session_c(self, lsn, nm1, nm2, peak):
        """ Save the output of session C analysis to self.save_path.  

        Parameters
        ----------
        lsn: LocallySparseNoise instance

        nm1: NaturalMovie instance
            This NaturalMovie instance should have been created with
            movie_name = stimulus_info.NATURAL_MOVIE_ONE

        nm2: NaturalMovie instance
            This NaturalMovie instance should have been created with
            movie_name = stimulus_info.NATURAL_MOVIE_TWO

        peak: pd.DataFrame
            The combined peak response property table created in self.session_c().
        """

        nwb = BrainObservatoryNwbDataSet(self.save_path)

        nwb.save_analysis_dataframes(
            ('stim_table_lsn', lsn.stim_table),
            ('sweep_response_nm1', nm1.sweep_response),
            ('peak', peak),
            ('sweep_response_nm2', nm2.sweep_response),
            ('sweep_response_lsn', lsn.sweep_response),
            ('mean_sweep_response_lsn', lsn.mean_sweep_response))

        nwb.save_analysis_arrays(
            ('receptive_field_lsn', lsn.receptive_field),
            ('mean_response_lsn', lsn.mean_response),
            ('binned_dx_sp', nm1.binned_dx_sp),
            ('binned_dx_vis', nm1.binned_dx_vis),
            ('binned_cells_sp', nm1.binned_cells_sp),
            ('binned_cells_vis', nm1.binned_cells_vis))

        LocallySparseNoise.save_cell_index_receptive_field_analysis(lsn.cell_index_receptive_field_analysis_data, nwb, stimulus_info.LOCALLY_SPARSE_NOISE)

    def save_session_c2(self, lsn4, lsn8, nm1, nm2, peak):        
        """ Save the output of session C2 analysis to self.save_path. 

        Parameters
        ----------
        lsn4: LocallySparseNoise instance
            This LocallySparseNoise instance should have been created with 
            self.stimulus = stimulus_info.LOCALLY_SPARSE_NOISE_4DEG.

        lsn8: LocallySparseNoise instance
            This LocallySparseNoise instance should have been created with 
            self.stimulus = stimulus_info.LOCALLY_SPARSE_NOISE_8DEG.

        nm1: NaturalMovie instance
            This NaturalMovie instance should have been created with
            movie_name = stimulus_info.NATURAL_MOVIE_ONE

        nm2: NaturalMovie instance
            This NaturalMovie instance should have been created with
            movie_name = stimulus_info.NATURAL_MOVIE_TWO

        peak: pd.DataFrame
            The combined peak response property table created in self.session_c2().
        """

        nwb = BrainObservatoryNwbDataSet(self.save_path)

        nwb.save_analysis_dataframes(
            ('stim_table_lsn4', lsn4.stim_table),
            ('stim_table_lsn8', lsn8.stim_table),
            ('sweep_response_nm1', nm1.sweep_response),
            ('peak', peak),
            ('sweep_response_nm2', nm2.sweep_response),
            ('sweep_response_lsn4', lsn4.sweep_response),
            ('sweep_response_lsn8', lsn8.sweep_response),
            ('mean_sweep_response_lsn4', lsn4.mean_sweep_response),
            ('mean_sweep_response_lsn8', lsn8.mean_sweep_response))

        merge_mean_response = LocallySparseNoise.merge_mean_response(
            lsn4.mean_response,
            lsn8.mean_response)

        nwb.save_analysis_arrays(
            ('mean_response_lsn4', lsn4.mean_response),
            ('mean_response_lsn8', lsn8.mean_response),
            ('receptive_field_lsn4', lsn4.receptive_field),
            ('receptive_field_lsn8', lsn8.receptive_field),
            ('merge_mean_response', merge_mean_response),
            ('binned_dx_sp', nm1.binned_dx_sp),
            ('binned_dx_vis', nm1.binned_dx_vis),
            ('binned_cells_sp', nm1.binned_cells_sp),
            ('binned_cells_vis', nm1.binned_cells_vis))

        LocallySparseNoise.save_cell_index_receptive_field_analysis(lsn4.cell_index_receptive_field_analysis_data, nwb, stimulus_info.LOCALLY_SPARSE_NOISE_4DEG)
        LocallySparseNoise.save_cell_index_receptive_field_analysis(lsn8.cell_index_receptive_field_analysis_data, nwb, stimulus_info.LOCALLY_SPARSE_NOISE_8DEG)

    def append_metrics_drifting_grating(self, metrics, dg):
        """ Extract metrics from the DriftingGratings peak response table into a dictionary. """

        metrics["osi_dg"] = dg.peak["osi_dg"]
        metrics["dsi_dg"] = dg.peak["dsi_dg"]
        metrics["pref_dir_dg"] = [dg.orivals[i]
                                  for i in dg.peak["ori_dg"].values]
        metrics["pref_tf_dg"] = [dg.tfvals[i] for i in dg.peak["tf_dg"].values]
        metrics["p_dg"] = dg.peak["ptest_dg"]
        metrics["g_osi_dg"] = dg.peak["cv_os_dg"]
        metrics["g_dsi_dg"] = dg.peak["cv_ds_dg"]
        metrics["reliability_dg"] = dg.peak["reliability_dg"]
        metrics["tfdi_dg"] = dg.peak["tf_index_dg"]
        metrics["run_mod_dg"] = dg.peak["run_modulation_dg"]
        metrics["p_run_mod_dg"] = dg.peak["p_run_dg"]
        metrics["peak_dff_dg"] = dg.peak["peak_dff_dg"]

    def append_metrics_static_grating(self, metrics, sg):
        """ Extract metrics from the StaticGratings peak response table into a dictionary. """

        metrics["osi_sg"] = sg.peak["osi_sg"]
        metrics["pref_ori_sg"] = [sg.orivals[i]
                                  for i in sg.peak["ori_sg"].values]
        metrics["pref_sf_sg"] = [sg.sfvals[i] for i in sg.peak["sf_sg"].values]
        metrics["pref_phase_sg"] = [sg.phasevals[i]
                                    for i in sg.peak["phase_sg"].values]
        metrics["p_sg"] = sg.peak["ptest_sg"]
        metrics["time_to_peak_sg"] = sg.peak["time_to_peak_sg"]
        metrics["run_mod_sg"] = sg.peak["run_modulation_sg"]
        metrics["p_run_mod_sg"] = sg.peak["p_run_sg"]
        metrics["g_osi_sg"] = sg.peak["cv_os_sg"]
        metrics["sfdi_sg"] = sg.peak["sf_index_sg"]
        metrics["peak_dff_sg"] = sg.peak["peak_dff_sg"]
        metrics["reliability_sg"] = sg.peak["reliability_sg"]

    def append_metrics_natural_scene(self, metrics, ns):
        """ Extract metrics from the NaturalScenes peak response table into a dictionary. """

        metrics["pref_image_ns"] = ns.peak["scene_ns"]
        metrics["p_ns"] = ns.peak["ptest_ns"]
        metrics["time_to_peak_ns"] = ns.peak["time_to_peak_ns"]
        metrics["image_sel_ns"] = ns.peak["image_selectivity_ns"]
        metrics["reliability_ns"] = ns.peak["reliability_ns"]
        metrics["run_mod_ns"] = ns.peak["run_modulation_ns"]
        metrics["p_run_mod_ns"] = ns.peak["p_run_ns"]
        metrics["peak_dff_ns"] = ns.peak["peak_dff_ns"]

    def append_metrics_locally_sparse_noise(self, metrics, lsn):
        """ Extract metrics from the LocallySparseNoise peak response table into a dictionary. """

        metrics['rf_chi2_lsn'] = lsn.peak['rf_chi2_lsn']
        metrics['rf_area_on_lsn'] = lsn.peak['rf_area_on_lsn']
        metrics['rf_center_on_x_lsn'] = lsn.peak['rf_center_on_x_lsn']
        metrics['rf_center_on_y_lsn'] = lsn.peak['rf_center_on_y_lsn']
        metrics['rf_area_off_lsn'] = lsn.peak['rf_area_off_lsn']
        metrics['rf_center_off_x_lsn'] = lsn.peak['rf_center_off_x_lsn']
        metrics['rf_center_off_y_lsn'] = lsn.peak['rf_center_off_y_lsn']
        metrics['rf_distance_lsn'] = lsn.peak['rf_distance_lsn']
        metrics['rf_overlap_index_lsn'] = lsn.peak['rf_overlap_index_lsn']

    def append_metrics_natural_movie_one(self, metrics, nma):
        """ Extract metrics from the NaturalMovie(stimulus_info.NATURAL_MOVIE_ONE) peak response table into a dictionary. """
        metrics['reliability_nm1'] = nma.peak['response_reliability_nm1']

    def append_metrics_natural_movie_two(self, metrics, nma):
        """ Extract metrics from the NaturalMovie(stimulus_info.NATURAL_MOVIE_TWO) peak response table into a dictionary. """
        metrics['reliability_nm2'] = nma.peak['response_reliability_nm2']

    def append_metrics_natural_movie_three(self, metrics, nma):
        """ Extract metrics from the NaturalMovie(stimulus_info.NATURAL_MOVIE_THREE) peak response table into a dictionary. """
        metrics['reliability_nm3'] = nma.peak['response_reliability_nm3']

    def append_experiment_metrics(self, metrics):
        """ Extract stimulus-agnostic metrics from an experiment into a dictionary """
        dxcm, dxtime = self.nwb.get_running_speed()
        metrics['mean_running_speed'] = np.nanmean(dxcm)

    def verify_roi_lists_equal(self, roi1, roi2):
        """ TODO: replace this with simpler numpy comparisons """

        if len(roi1) != len(roi2):
            raise BrainObservatoryAnalysisException(
                "Error -- ROI lists are of different length")

        for i in range(len(roi1)):
            if roi1[i] != roi2[i]:
                raise BrainObservatoryAnalysisException(
                    "Error -- ROI lists have different entries")

    def session_a(self, plot_flag=False, save_flag=True):
        """ Run stimulus-specific analysis for natural movie one, natural movie three, and drifting gratings.
        The input NWB be for a stimulus_info.THREE_SESSION_A experiment.

        Parameters
        ----------
        plot_flag: bool
            Whether to generate brain_observatory_plotting work plots after running analysis.

        save_flag: bool
            Whether to save the output of analysis to self.save_path upon completion.
        """

        nm1 = NaturalMovie(self.nwb, 'natural_movie_one')
        nm3 = NaturalMovie(self.nwb, 'natural_movie_three')
        dg = DriftingGratings(self.nwb)

        dg.noise_correlation, _, _, _ = dg.get_noise_correlation()
        dg.signal_correlation, _ = dg.get_signal_correlation()
        dg.representational_similarity, _ = dg.get_representational_similarity()

        SessionAnalysis._log.info("Session A analyzed")
        peak = multi_dataframe_merge(
            [nm1.peak_run, dg.peak, nm1.peak, nm3.peak])

        self.append_metrics_drifting_grating(self.metrics_a['cell'], dg)
        self.append_metrics_natural_movie_one(self.metrics_a['cell'], nm1)
        self.append_metrics_natural_movie_three(self.metrics_a['cell'], nm3)
        self.append_experiment_metrics(self.metrics_a['experiment'])
        self.metrics_a['cell']['roi_id'] = dg.roi_id

        self.append_metadata(peak)

        if save_flag:
            self.save_session_a(dg, nm1, nm3, peak)

        if plot_flag:
            cp._plot_3sa(dg, nm1, nm3, self.save_dir)
            cp.plot_drifting_grating_traces(dg, self.save_dir)

    def session_b(self, plot_flag=False, save_flag=True):
        """ Run stimulus-specific analysis for natural scenes, static gratings, and natural movie one.
        The input NWB be for a stimulus_info.THREE_SESSION_B experiment.

        Parameters
        ----------
        plot_flag: bool
            Whether to generate brain_observatory_plotting work plots after running analysis.

        save_flag: bool
            Whether to save the output of analysis to self.save_path upon completion.
        """

        ns = NaturalScenes(self.nwb)
        sg = StaticGratings(self.nwb)
        nm1 = NaturalMovie(self.nwb, 'natural_movie_one')
        SessionAnalysis._log.info("Session B analyzed")
        peak = multi_dataframe_merge(
            [nm1.peak_run, sg.peak, ns.peak, nm1.peak])
        self.append_metadata(peak)

        self.append_metrics_static_grating(self.metrics_b['cell'], sg)
        self.append_metrics_natural_scene(self.metrics_b['cell'], ns)
        self.append_metrics_natural_movie_one(self.metrics_b['cell'], nm1)
        self.append_experiment_metrics(self.metrics_b['experiment'])
        self.verify_roi_lists_equal(sg.roi_id, ns.roi_id)
        self.metrics_b['cell']['roi_id'] = sg.roi_id

        sg.noise_correlation, _, _, _ = sg.get_noise_correlation()
        sg.signal_correlation, _ = sg.get_signal_correlation()
        sg.representational_similarity, _ = sg.get_representational_similarity()

        ns.noise_correlation, _ = ns.get_noise_correlation()
        ns.signal_correlation, _ = ns.get_signal_correlation()
        ns.representational_similarity, _ = ns.get_representational_similarity()

        if save_flag:
            self.save_session_b(sg, nm1, ns, peak)

        if plot_flag:
            cp._plot_3sb(sg, nm1, ns, self.save_dir)
            cp.plot_ns_traces(ns, self.save_dir)
            cp.plot_sg_traces(sg, self.save_dir)

    def session_c(self, plot_flag=False, save_flag=True):
        """ Run stimulus-specific analysis for natural movie one, natural movie two, and locally sparse noise.
        The input NWB be for a stimulus_info.THREE_SESSION_C experiment.

        Parameters
        ----------
        plot_flag: bool
            Whether to generate brain_observatory_plotting work plots after running analysis.

        save_flag: bool
            Whether to save the output of analysis to self.save_path upon completion.
        """

        lsn = LocallySparseNoise(self.nwb, stimulus_info.LOCALLY_SPARSE_NOISE)
        nm2 = NaturalMovie(self.nwb, 'natural_movie_two')
        nm1 = NaturalMovie(self.nwb, 'natural_movie_one')
        SessionAnalysis._log.info("Session C analyzed")
        peak = multi_dataframe_merge([nm1.peak_run, nm1.peak, nm2.peak, lsn.peak])
        self.append_metadata(peak)

        self.append_metrics_locally_sparse_noise(self.metrics_c['cell'], lsn)
        self.append_metrics_natural_movie_one(self.metrics_c['cell'], nm1)
        self.append_metrics_natural_movie_two(self.metrics_c['cell'], nm2)
        self.append_experiment_metrics(self.metrics_c['experiment'])
        self.metrics_c['cell']['roi_id'] = nm1.roi_id

        if save_flag:
            self.save_session_c(lsn, nm1, nm2, peak)

        if plot_flag:
            cp._plot_3sc(lsn, nm1, nm2, self.save_dir)
            cp.plot_lsn_traces(lsn, self.save_dir)

    def session_c2(self, plot_flag=False, save_flag=True):
        """ Run stimulus-specific analysis for locally sparse noise (4 deg.), locally sparse noise (8 deg.),
        natural movie one, and natural movie two. The input NWB be for a stimulus_info.THREE_SESSION_C2 experiment.

        Parameters
        ----------
        plot_flag: bool
            Whether to generate brain_observatory_plotting work plots after running analysis.

        save_flag: bool
            Whether to save the output of analysis to self.save_path upon completion.
        """

        lsn4 = LocallySparseNoise(self.nwb, stimulus_info.LOCALLY_SPARSE_NOISE_4DEG)
        lsn8 = LocallySparseNoise(self.nwb, stimulus_info.LOCALLY_SPARSE_NOISE_8DEG)

        nm2 = NaturalMovie(self.nwb, 'natural_movie_two')
        nm1 = NaturalMovie(self.nwb, 'natural_movie_one')
        SessionAnalysis._log.info("Session C2 analyzed")

        if self.nwb.get_metadata()['targeted_structure'] == 'VISp':
            lsn_peak = lsn4
        else:
            lsn_peak = lsn8

        peak = multi_dataframe_merge([nm1.peak_run, nm1.peak, nm2.peak, lsn_peak.peak])
        self.append_metadata(peak)

        self.append_metrics_locally_sparse_noise(self.metrics_c['cell'], lsn_peak)
        self.append_metrics_natural_movie_one(self.metrics_c['cell'], nm1)
        self.append_metrics_natural_movie_two(self.metrics_c['cell'], nm2)
        self.append_experiment_metrics(self.metrics_c['experiment'])
        self.metrics_c['cell']['roi_id'] = nm1.roi_id

        if save_flag:
            self.save_session_c2(lsn4, lsn8, nm1, nm2, peak)

        if plot_flag:
            cp._plot_3sc(lsn4, nm1, nm2, self.save_dir, '_4deg')
            cp._plot_3sc(lsn8, nm1, nm2, self.save_dir, '_8deg')
            cp.plot_lsn_traces(lsn4, self.save_dir, '_4deg')
            cp.plot_lsn_traces(lsn4, self.save_dir, '_8deg')


def run_session_analysis(nwb_path, save_path, plot_flag=False, save_flag=True):
    """ Inspect an NWB file to determine which experiment session was run
    and compute all stimulus-specific analyses.

    Parameters
    ----------
    nwb_path: string
        Path to NWB file.

    save_path: string
        path to save results. Recommended NOT to use NWB file.

    plot_flag: bool
        Whether to save brain_observatory_plotting work plots.

    save_flag: bool
        Whether to save results to save_path.
    """

    save_dir = os.path.abspath(os.path.dirname(save_path))

    if not os.path.exists(save_dir):
        os.makedirs(save_dir)

    session_analysis = SessionAnalysis(nwb_path, save_path)

    session = session_analysis.nwb.get_session_type()

    if session == stimulus_info.THREE_SESSION_A:
        session_analysis.session_a(plot_flag=plot_flag, save_flag=save_flag)
        metrics = session_analysis.metrics_a
    elif session == stimulus_info.THREE_SESSION_B:
        session_analysis.session_b(plot_flag=plot_flag, save_flag=save_flag)
        metrics = session_analysis.metrics_b
    elif session == stimulus_info.THREE_SESSION_C:
        session_analysis.session_c(plot_flag=plot_flag, save_flag=save_flag)
        metrics = session_analysis.metrics_c
    elif session == stimulus_info.THREE_SESSION_C2:
        session_analysis.session_c2(plot_flag=plot_flag, save_flag=save_flag)
        metrics = session_analysis.metrics_c
    else:
        raise IndexError("Unknown session: %s" % session)

    return metrics


def main():
    parser = argparse.ArgumentParser()
    parser.add_argument("input_nwb")
    parser.add_argument("output_h5")

    parser.add_argument("--plot", action='store_true')

    args = parser.parse_args()
    logging.basicConfig()
    logging.getLogger().setLevel(logging.DEBUG)

    run_session_analysis(args.input_nwb, args.output_h5, args.plot)


if __name__ == '__main__':
    main()
